import os
import shutil
from pathlib import Path
from typing import Dict, List, Tuple

import pandas as pd

from mlebench.utils import extract, read_csv


def filter_and_write_file(src: Path, dst: Path, old_id_to_new: Dict[int, int]):
    """
    Given txt file that has column 0 as rec_id, filters out rec_ids that are not in old_id_to_new and writes to dst
    """
    history_of_segments = open(src).read().splitlines()
    history_of_segments = history_of_segments[1:]
    history_of_segments = [
        (int(i.split(",")[0]), ",".join(i.split(",")[1:])) for i in history_of_segments
    ]
    history_of_segments = [
        (old_id_to_new[i[0]], i[1]) for i in history_of_segments if i[0] in old_id_to_new.keys()
    ]
    with open(dst, "w") as f:
        f.write("rec_id,[histogram of segment features]\n")
        for rec_id, labels in history_of_segments:
            f.write(f"{rec_id},{labels}\n")


def _process_split(
    cv_folds: pd.DataFrame,
    full_rec_id2filename: pd.DataFrame,
    full_rec_labels_split: List[Tuple[int, str]],
    raw_path: Path,
    public_path: Path,
    private_path: Path,
):
    """
    Helper function to process a single data split and generate corresponding public and private files.
    """
    public_path.mkdir(exist_ok=True, parents=True)
    private_path.mkdir(exist_ok=True, parents=True)
    (public_path / "essential_data").mkdir(exist_ok=True)
    (public_path / "supplemental_data").mkdir(exist_ok=True)

    # Create a new compact ID mapping for the current subset of data
    old_ids_in_split = cv_folds["rec_id"].values
    old_id_to_new = {old_id: new_id for new_id, old_id in enumerate(old_ids_in_split)}
    cv_folds_processed = cv_folds.copy()
    cv_folds_processed["rec_id"] = cv_folds_processed["rec_id"].map(old_id_to_new)
    cv_folds_processed.to_csv(public_path / "essential_data/CVfolds_2.txt", index=False)

    test_rec_ids = cv_folds_processed[cv_folds_processed["fold"] == 1]["rec_id"].values

    # Update id2filename with new split
    rec_id2filename = full_rec_id2filename[
        full_rec_id2filename["rec_id"].isin(old_id_to_new.keys())
    ].copy()
    rec_id2filename["rec_id"] = rec_id2filename["rec_id"].map(old_id_to_new)
    rec_id2filename.to_csv(public_path / "essential_data/rec_id2filename.txt", index=False)
    assert len(rec_id2filename) == len(
        cv_folds_processed
    ), f"Expected {len(cv_folds_processed)} entires in rec_id2filename, got {len(rec_id2filename)}"

    # Update labels with new split
    rec_labels_split = [i for i in full_rec_labels_split if i[0] in old_id_to_new.keys()]
    rec_labels_split = [(old_id_to_new[i[0]], i[1]) for i in rec_labels_split]

    # Public labels
    with open(public_path / "essential_data/rec_labels_test_hidden.txt", "w") as f:
        f.write("rec_id,[labels]\n")
        for rec_id, labels in rec_labels_split:
            if rec_id in test_rec_ids:
                labels = "?"
            if labels == "":  # Write without comma
                f.write(f"{rec_id}{labels}\n")
            else:
                f.write(f"{rec_id},{labels}\n")

    # Private labels. Create csv, with each row containing the label for a (rec_id, species_id) pair
    data = {"Id": [], "Probability": []}
    for rec_id, labels in rec_labels_split:
        if rec_id not in test_rec_ids:
            continue
        species_ids = [int(i) for i in labels.split(",") if i != ""]
        for species_id in range(0, 19):
            data["Id"].append(rec_id * 100 + species_id)
            data["Probability"].append(int(species_id in species_ids))

    pd.DataFrame(data).to_csv(private_path / "answers.csv", index=False)
    if len(test_rec_ids) > 0:
        assert (
            len(pd.DataFrame(data)) == len(test_rec_ids) * 19
        ), f"Expected {len(test_rec_ids)*19} entires in answers.csv, got {len(pd.DataFrame(data))}"

    # Create new sample submission, following new submission format
    data = {
        "Id": [rec_id * 100 + species_id for rec_id in test_rec_ids for species_id in range(0, 19)],
        "Probability": 0,
    }
    pd.DataFrame(data).to_csv(public_path / "sample_submission.csv", index=False)
    if len(test_rec_ids) > 0:
        assert (
            len(pd.DataFrame(data)) == len(test_rec_ids) * 19
        ), f"Expected {len(test_rec_ids)*19} entires in sample_submission.csv, got {len(pd.DataFrame(data))}"

    # Copy over species list
    shutil.copyfile(
        src=raw_path / "mlsp_contest_dataset/essential_data/species_list.txt",
        dst=public_path / "essential_data/species_list.txt",
    )

    # Copy over all src waves from train+test set
    (public_path / "essential_data/src_wavs").mkdir(exist_ok=True)
    for filename in rec_id2filename["filename"]:
        shutil.copyfile(
            src=raw_path / "mlsp_contest_dataset/essential_data/src_wavs" / f"{filename}.wav",
            dst=public_path / "essential_data/src_wavs" / f"{filename}.wav",
        )

    # Copy over train+test filtered spectrograms, segmentation examples, spectrograms, and supervised segmentation
    (public_path / "supplemental_data/filtered_spectrograms").mkdir(exist_ok=True)
    (public_path / "supplemental_data/segmentation_examples").mkdir(exist_ok=True)
    (public_path / "supplemental_data/spectrograms").mkdir(exist_ok=True)
    (public_path / "supplemental_data/supervised_segmentation").mkdir(exist_ok=True)
    for filename in rec_id2filename["filename"]:
        shutil.copyfile(
            src=raw_path
            / "mlsp_contest_dataset/supplemental_data/filtered_spectrograms"
            / f"{filename}.bmp",
            dst=public_path / "supplemental_data/filtered_spectrograms" / f"{filename}.bmp",
        )
        if os.path.exists(
            raw_path / "mlsp_contest_dataset/supplemental_data/segmentation_examples" / f"{filename}.bmp"
        ):
            shutil.copyfile(
                src=raw_path
                / "mlsp_contest_dataset/supplemental_data/segmentation_examples"
                / f"{filename}.bmp",
                dst=public_path / "supplemental_data/segmentation_examples" / f"{filename}.bmp",
            )
        shutil.copyfile(
            src=raw_path / "mlsp_contest_dataset/supplemental_data/spectrograms" / f"{filename}.bmp",
            dst=public_path / "supplemental_data/spectrograms" / f"{filename}.bmp",
        )
        shutil.copyfile(
            src=raw_path
            / "mlsp_contest_dataset/supplemental_data/supervised_segmentation"
            / f"{filename}.bmp",
            dst=public_path / "supplemental_data/supervised_segmentation" / f"{filename}.bmp",
        )

    # Copy over remaining files
    shutil.copyfile(
        src=raw_path / "mlsp_contest_dataset/supplemental_data/segment_clusters.bmp",
        dst=public_path / "supplemental_data/segment_clusters.bmp",
    )
    shutil.copyfile(
        src=raw_path / "mlsp_contest_dataset/supplemental_data/segment_mosaic.bmp",
        dst=public_path / "supplemental_data/segment_mosaic.bmp",
    )

    filter_and_write_file(
        src=raw_path / "mlsp_contest_dataset/supplemental_data/histogram_of_segments.txt",
        dst=public_path / "supplemental_data/histogram_of_segments.txt",
        old_id_to_new=old_id_to_new,
    )
    filter_and_write_file(
        src=raw_path / "mlsp_contest_dataset/supplemental_data/segment_features.txt",
        dst=public_path / "supplemental_data/segment_features.txt",
        old_id_to_new=old_id_to_new,
    )
    filter_and_write_file(
        src=raw_path / "mlsp_contest_dataset/supplemental_data/segment_rectangles.txt",
        dst=public_path / "supplemental_data/segment_rectangles.txt",
        old_id_to_new=old_id_to_new,
    )


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split in public_val and private_val directories.
    """
    # extract only what we need
    extract(raw / "mlsp_contest_dataset.zip", raw)

    # --- Load all necessary data once ---
    full_cv_folds = read_csv(raw / "mlsp_contest_dataset/essential_data/CVfolds_2.txt")
    full_rec_id2filename = read_csv(raw / "mlsp_contest_dataset/essential_data/rec_id2filename.txt")
    rec_labels_raw = (
        open(raw / "mlsp_contest_dataset/essential_data/rec_labels_test_hidden.txt")
        .read()
        .splitlines()
    )
    rec_labels_raw = rec_labels_raw[1:]  # Ignore header line
    full_rec_labels_split = []
    for i in rec_labels_raw:
        rec_id = i.split(",")[0]
        labels = ",".join(i.split(",")[1:]) if len(i.split(",")) > 1 else ""
        full_rec_labels_split.append((int(rec_id), labels))

    # --- 1. Create the original Train/Test split ---
    # Filter to only the data used in the original split logic and create test set
    cv_folds = full_cv_folds[full_cv_folds["fold"] == 0].reset_index(drop=True)
    test_indices = cv_folds.sample(frac=0.2, random_state=0).index
    cv_folds.loc[test_indices, "fold"] = 1
    assert len(test_indices) == 64, f"Expected 64 test rec_ids, got {len(test_indices)}"

    # Process and write files for the original public/private split
    _process_split(
        cv_folds=cv_folds,
        full_rec_id2filename=full_rec_id2filename,
        full_rec_labels_split=full_rec_labels_split,
        raw_path=raw,
        public_path=public,
        private_path=private,
    )

    # --- 2. Create the new Train/Validation split from the first split's TRAIN set ---
    # Define new paths for the validation split
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Take the training data from the first split (fold=0)
    train_from_first_split = cv_folds[cv_folds["fold"] == 0].reset_index(drop=True)

    # Re-split this training data to create a new train_val and test_val set
    # The new test (validation) set should be ~64 samples.
    # Original train size = 256. 64/256 = 0.25
    val_indices = train_from_first_split.sample(frac=0.25, random_state=0).index
    train_from_first_split.loc[val_indices, "fold"] = 1  # Re-use 'fold=1' to mark test set

    # Process and write files for the new validation split
    _process_split(
        cv_folds=train_from_first_split,
        full_rec_id2filename=full_rec_id2filename,
        full_rec_labels_split=full_rec_labels_split,
        raw_path=raw,
        public_path=public_val,
        private_path=private_val,
    )


import pandas as pd
from sklearn.metrics import f1_score

from mlebench.grade_helpers import InvalidSubmissionError


def prepare_for_metric(submission: pd.DataFrame, answers: dict):
    """
    `submission` is a pd.DataFrame with "example_id" and "PredictionString" columns.

    `answers` is a dict in the format of "simplified-nq-train.jsonl" described here:
    https://www.kaggle.com/c/tensorflow2-question-answering/data

    The competition uses a micro F1 score metric, which is a binary classification metric.
    For retrieval, we convert the submission and answers for each sample as:
    - y_pred: if the submission has an answer, it is 1, else 0
    - y_true:
        - if the submission has an answer and it exists in the true labels, y_true=1
        - if the submission has an answer but it does not exist in the true labels, y_true=0
        - if the submission has no answer but there exists a true label, y_true=1
        - if the submission has no answer and there is no true label, y_true=0

    This is consistent with
    https://www.kaggle.com/competitions/tensorflow2-question-answering/overview/evaluation
    - TP = the predicted indices match one of the possible ground truth indices
    - FP = the predicted indices do NOT match one of the possible ground truth indices, OR
        a prediction has been made where no ground truth exists
    - FN = no prediction has been made where a ground truth exists

    Returns y_true, y_pred which are lists of 0s and 1s.
    """
    if len(submission) != 2 * len(answers):
        raise InvalidSubmissionError(
            f"Submission length {len(submission)} != 2 * answers length {len(answers)}"
        )
    # Empty strings are read as NaN by pandas, but we want these to remain empty strings
    submission.fillna("", inplace=True)
    submission = submission.astype(str)

    expected_ids = []
    for sample in answers:
        expected_ids.append(f"{sample['example_id']}_long")
        expected_ids.append(f"{sample['example_id']}_short")

    if not ({"example_id", "PredictionString"}).issubset(submission.columns):
        raise InvalidSubmissionError(
            "Submission requires 'example_id' and 'PredictionString' columns."
        )
    if not set(submission["example_id"]) == set(expected_ids):
        raise InvalidSubmissionError(
            "Submission example_ids do not match expected_ids. Please ensure you have a "
            "long and short answer for each example_id in the answers."
        )
    if not len(set(submission["example_id"])) == len(submission):
        raise InvalidSubmissionError(
            "Submission example_ids are not unique. Please ensure you have a "
            "long and short answer for each example_id in the answers."
        )

    y_pred = [1 if row["PredictionString"].strip() else 0 for _, row in submission.iterrows()]

    y_true = []
    true_annotations = {sample["example_id"]: sample["annotations"][0] for sample in answers}
    for idx, row in submission.iterrows():
        sample_id = row["example_id"]

        # Parse prediction
        # Prediction may be any of "{start_token}:{end_token}", "YES", "NO", ""
        pred = row["PredictionString"].strip()
        if ":" in pred and len(pred.split(":")) == 2:
            start_token, end_token = pred.split(":")
            pred = (int(start_token), int(end_token))
        elif pred in ["YES", "NO", ""]:
            pass
        else:
            raise InvalidSubmissionError(f"Invalid submission format: {pred}")

        # Parse valid answers
        if sample_id.endswith("_long"):
            sample_id = sample_id.split("_long")[0]
            annotation = true_annotations[sample_id]
            valid_answers = []
            if annotation["long_answer"]["start_token"] != -1:
                valid_answers.append(
                    (
                        int(annotation["long_answer"]["start_token"]),
                        int(annotation["long_answer"]["end_token"]),
                    ),
                )
        elif sample_id.endswith("_short"):
            sample_id = sample_id.split("_short")[0]
            annotation = true_annotations[sample_id]
            valid_answers = [
                (int(short_answer["start_token"]), int(short_answer["end_token"]))
                for short_answer in annotation["short_answers"]
            ]
            if annotation["yes_no_answer"] != "NONE":
                valid_answers.append(annotation["yes_no_answer"])
        else:
            raise InvalidSubmissionError(
                f"sample_id {sample_id} does not end with '_long' or '_short'"
            )

        # Mark as 0 or 1 based on the above rules
        if pred:
            if pred in valid_answers:
                y_true.append(1)
            else:
                y_true.append(0)
        else:
            if valid_answers:
                y_true.append(1)
            else:
                y_true.append(0)

    assert len(y_true) == len(y_pred), f"y_true {len(y_true)} != y_pred {len(y_pred)}"
    return y_true, y_pred


def grade(submission: pd.DataFrame, answers: dict) -> float:
    y_true, y_pred = prepare_for_metric(submission, answers)
    return f1_score(y_true=y_true, y_pred=y_pred, average="micro")



import argparse
import json
from pathlib import Path

from mlebench.data import download_and_prepare_dataset, ensure_leaderboard_exists
from mlebench.grade import grade_csv, grade_jsonl
from mlebench.registry import registry
from mlebench.utils import get_logger

logger = get_logger(__name__)


def main():
    parser = argparse.ArgumentParser(description="Runs agents on Kaggle competitions.")
    subparsers = parser.add_subparsers(dest="command", help="Sub-command to run.")

    # Prepare sub-parser
    parser_prepare = subparsers.add_parser(
        name="prepare",
        help="Download and prepare competitions for the MLE-bench dataset.",
    )
    parser_prepare.add_argument(
        "-c",
        "--competition-id",
        help=f"ID of the competition to prepare. Valid options: {registry.list_competition_ids()}",
        type=str,
        required=False,
    )
    parser_prepare.add_argument(
        "-a",
        "--all",
        help="Prepare all competitions.",
        action="store_true",
    )
    parser_prepare.add_argument(
        "--lite",
        help="Prepare all the low complexity competitions (MLE-bench Lite).",
        action="store_true",
        required=False,
    )
    parser_prepare.add_argument(
        "-l",
        "--list",
        help="Prepare a list of competitions specified line by line in a text file.",
        type=str,
        required=False,
    )
    parser_prepare.add_argument(
        "--keep-raw",
        help="Keep the raw competition files after the competition has been prepared.",
        action="store_true",
        required=False,
        default=False,
    )
    parser_prepare.add_argument(
        "--data-dir",
        help="Path to the directory where the data will be stored.",
        required=False,
        default=registry.get_data_dir(),
    )
    parser_prepare.add_argument(
        "--overwrite-checksums",
        help="[For Developers] Overwrite the checksums file for the competition.",
        action="store_true",
        required=False,
        default=False,
    )
    parser_prepare.add_argument(
        "--overwrite-leaderboard",
        help="[For Developers] Overwrite the leaderboard file for the competition.",
        action="store_true",
        required=False,
        default=False,
    )
    parser_prepare.add_argument(
        "--skip-verification",
        help="[For Developers] Skip the verification of the checksums.",
        action="store_true",
        required=False,
        default=True,
    )

    # Grade eval sub-parser
    parser_grade_eval = subparsers.add_parser(
        "grade",
        help="Grade a submission to the eval, comprising of several competition submissions",
    )
    parser_grade_eval.add_argument(
        "--submission",
        help="Path to the JSONL file of submissions. Refer to README.md#submission-format for the required format.",
        type=str,
        required=True,
    )
    parser_grade_eval.add_argument(
        "--output-dir",
        help="Path to the directory where the evaluation metrics will be saved.",
        type=str,
        required=True,
    )
    parser_grade_eval.add_argument(
        "--data-dir",
        help="Path to the directory where the data used for grading is stored.",
        required=False,
        default=registry.get_data_dir(),
    )

    # Grade sample sub-parser
    parser_grade_sample = subparsers.add_parser(
        name="grade-sample",
        help="Grade a single sample (competition) in the eval",
    )
    parser_grade_sample.add_argument(
        "submission",
        help="Path to the submission CSV file.",
        type=str,
    )
    parser_grade_sample.add_argument(
        "competition_id",
        help=f"ID of the competition to grade. Valid options: {registry.list_competition_ids()}",
        type=str,
    )
    parser_grade_sample.add_argument(
        "--data-dir",
        help="Path to the directory where the data will be stored.",
        required=False,
        default=registry.get_data_dir(),
    )

    # Dev tools sub-parser
    parser_dev = subparsers.add_parser("dev", help="Developer tools for extending MLE-bench.")
    dev_subparsers = parser_dev.add_subparsers(dest="dev_command", help="Developer command to run.")

    # Set up 'download-leaderboard' under 'dev'
    parser_download_leaderboard = dev_subparsers.add_parser(
        "download-leaderboard",
        help="Download the leaderboard for a competition.",
    )
    parser_download_leaderboard.add_argument(
        "-c",
        "--competition-id",
        help=f"Name of the competition to download the leaderboard for. Valid options: {registry.list_competition_ids()}",
        type=str,
        required=False,
    )
    parser_download_leaderboard.add_argument(
        "--all",
        help="Download the leaderboard for all competitions.",
        action="store_true",
    )
    parser_download_leaderboard.add_argument(
        "--force",
        help="Force download the leaderboard, even if it already exists.",
        action="store_true",
    )

    args = parser.parse_args()

    if args.command == "prepare":
        new_registry = registry.set_data_dir(Path(args.data_dir))
        new_registry.set_mode("prepare")

        if args.lite:
            competitions = [
                new_registry.get_competition(competition_id)
                for competition_id in new_registry.get_lite_competition_ids()
            ]
        elif args.all:
            competitions = [
                new_registry.get_competition(competition_id)
                for competition_id in registry.list_competition_ids()
            ]
        elif args.list:
            with open(args.list, "r") as f:
                competition_ids = f.read().splitlines()
            competitions = [
                new_registry.get_competition(competition_id) for competition_id in competition_ids
            ]
        else:
            if not args.competition_id:
                parser_prepare.error(
                    "One of --lite, --all, --list, or --competition-id must be specified."
                )
            competitions = [new_registry.get_competition(args.competition_id)]

        for competition in competitions:
            download_and_prepare_dataset(
                competition=competition,
                keep_raw=args.keep_raw,
                overwrite_checksums=args.overwrite_checksums,
                overwrite_leaderboard=args.overwrite_leaderboard,
                skip_verification=args.skip_verification,
            )
    if args.command == "grade":
        new_registry = registry.set_data_dir(Path(args.data_dir))
        submission = Path(args.submission)
        output_dir = Path(args.output_dir)
        grade_jsonl(submission, output_dir, new_registry)
    if args.command == "grade-sample":
        new_registry = registry.set_data_dir(Path(args.data_dir))
        competition = new_registry.get_competition(args.competition_id)
        submission = Path(args.submission)
        report = grade_csv(submission, competition)
        logger.info("Competition report:")
        logger.info(json.dumps(report.to_dict(), indent=4))
    if args.command == "dev":
        if args.dev_command == "download-leaderboard":
            if args.all:
                for competition_id in registry.list_competition_ids():
                    competition = registry.get_competition(competition_id)
                    ensure_leaderboard_exists(competition, force=args.force)
            elif args.competition_id:
                competition = registry.get_competition(args.competition_id)
                ensure_leaderboard_exists(competition, force=args.force)
            else:
                parser_download_leaderboard.error(
                    "Either --all or --competition-id must be specified."
                )


if __name__ == "__main__":
    main()



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import extract, read_csv


def _process_split(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    public_dir: Path,
    private_dir: Path,
    raw_images_dir: Path,
    raw_masks_dir: Path,
    all_depths_df: pd.DataFrame,
) -> None:
    """
    Processes a single data split (e.g., train/test or train_val/test_val),
    saving all required files and performing sanity checks.

    Args:
        train_df: DataFrame for the training set.
        test_df: DataFrame for the test set.
        public_dir: The public output directory for this split.
        private_dir: The private output directory for this split.
        raw_images_dir: Path to the directory containing all source images.
        raw_masks_dir: Path to the directory containing all source masks.
        all_depths_df: DataFrame containing depth information for all samples.
    """
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)

    train_ids = set(train_df["id"])
    test_ids = set(test_df["id"])

    assert train_ids.isdisjoint(test_ids), "`id` is not disjoint between train and test sets"

    train_df.sort_values(by="id").to_csv(public_dir / "train.csv", index=False)
    test_df.sort_values(by="id").to_csv(private_dir / "test.csv", index=False)

    all_source_images = set(raw_images_dir.glob("*.png"))
    train_imgs = set(img for img in all_source_images if img.stem in train_ids)
    test_imgs = set(img for img in all_source_images if img.stem in test_ids)

    assert train_imgs.isdisjoint(test_imgs), "Images are not disjoint between train and test sets"

    (public_dir / "train" / "images").mkdir(parents=True, exist_ok=True)
    (public_dir / "train" / "masks").mkdir(parents=True, exist_ok=True)

    for fpath in train_imgs:
        shutil.copyfile(src=fpath, dst=public_dir / "train" / "images" / fpath.name)
        shutil.copyfile(src=raw_masks_dir / fpath.name, dst=public_dir / "train" / "masks" / fpath.name)

    (public_dir / "test" / "images").mkdir(parents=True, exist_ok=True)

    for fpath in test_imgs:
        shutil.copyfile(src=fpath, dst=public_dir / "test" / "images" / fpath.name)

    sample_submission = test_df.drop(columns=["rle_mask"]).copy()
    sample_submission["rle_mask"] = "1 1"
    sample_submission.sort_values(by="id").to_csv(public_dir / "sample_submission.csv", index=False)

    depths_mask = all_depths_df["id"].isin(train_ids)
    filtered_depths = all_depths_df[depths_mask]
    filtered_depths.sort_values(by="id").to_csv(public_dir / "depths.csv", index=False)

    # Sanity checks
    assert (public_dir / "train.csv").exists(), "`train.csv` doesn't exist!"
    assert (public_dir / "sample_submission.csv").exists(), "`sample_submission.csv` doesn't exist!"
    assert (public_dir / "depths.csv").exists(), "`depths.csv` doesn't exist!"
    assert (public_dir / "train").exists(), "`train` directory doesn't exist!"
    assert (public_dir / "test").exists(), "`test` directory doesn't exist!"
    assert (private_dir / "test.csv").exists(), "`test.csv` doesn't exist!"

    actual_train_imgs = set(img.stem for img in (public_dir / "train" / "images").glob("*.png"))
    actual_train_masks = set(img.stem for img in (public_dir / "train" / "masks").glob("*.png"))

    assert len(actual_train_imgs) == len(train_df), "The number of images in the train set doesn't match!"
    assert len(actual_train_masks) == len(train_df), "The number of masks in the train set doesn't match!"

    for train_id in train_df["id"]:
        assert (public_dir / "train" / "images" / f"{train_id}.png").exists()
        assert (public_dir / "train" / "masks" / f"{train_id}.png").exists()

    actual_test_imgs = set(img.stem for img in (public_dir / "test" / "images").glob("*.png"))

    assert not (public_dir / "test" / "masks").exists(), f"Expected `{public_dir}/test/masks` to not exist, but it does!"
    assert len(actual_test_imgs) == len(test_df), "The number of images in the test set doesn't match!"

    for test_id in test_df["id"]:
        assert (public_dir / "test" / "images" / f"{test_id}.png").exists()
        assert not (public_dir / "test" / "masks" / f"{test_id}.png").exists()

    assert actual_train_imgs.isdisjoint(actual_test_imgs), "Image sets overlap!"

    actual_sample_submission = read_csv(public_dir / "sample_submission.csv")
    actual_test = read_csv(private_dir / "test.csv")

    assert len(actual_sample_submission) == len(actual_test), "Sample submission and test set lengths differ!"
    assert set(actual_sample_submission["id"]) == set(actual_test["id"]), "Sample submission and test set IDs differ!"
    assert len(actual_test_imgs) == len(actual_test), "Test image count and test set length differ!"
    assert set(actual_test["id"]) == actual_test_imgs, "Test set IDs and test images differ!"


def prepare(raw: Path, public: Path, private: Path) -> None:
    extract(raw / "competition_data.zip", raw)

    old_train = read_csv(raw / "competition_data" / "train.csv")
    old_train = old_train.fillna("")
    old_depths = read_csv(raw / "depths.csv")

    # Original ratio is Train set - 4,000 samples; Test set - ~18,000 samples (82% ratio)
    # We use a 0.25 ratio to get number of test samples into thousand OOM
    new_train, new_test = train_test_split(old_train, test_size=0.25, random_state=0)

    assert len(new_train) + len(new_test) == len(
        old_train
    ), "Some samples were lost when creating the new train and test sets!"

    # Create the new validation split from the `new_train` set.
    # To make `test_val` have the same size as `new_test` (25% of original),
    # we need to take 1/3 of `new_train` (since 1/3 * 75% = 25%).
    train_val, test_val = train_test_split(new_train, test_size=(1/3), random_state=0)

    assert len(train_val) + len(test_val) == len(
        new_train
    ), "Some samples were lost when creating the validation train and test sets!"

    # Define paths for raw images and new validation output directories
    raw_images_dir = raw / "competition_data" / "train" / "images"
    raw_masks_dir = raw / "competition_data" / "train" / "masks"
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Process the original split, saving to `public` and `private`
    # This block ensures the original outputs are not modified.
    _process_split(
        train_df=new_train,
        test_df=new_test,
        public_dir=public,
        private_dir=private,
        raw_images_dir=raw_images_dir,
        raw_masks_dir=raw_masks_dir,
        all_depths_df=old_depths,
    )

    # Process the new validation split, saving to `public_val` and `private_val`
    _process_split(
        train_df=train_val,
        test_df=test_val,
        public_dir=public_val,
        private_dir=private_val,
        raw_images_dir=raw_images_dir,
        raw_masks_dir=raw_masks_dir,
        all_depths_df=old_depths,
    )

    # Final checks on data types, which are consistent across all splits.
    assert new_train.applymap(
        lambda x: isinstance(x, str)
    ).values.all(), "Not all elements in the DataFrame are strings!"
    assert new_test.applymap(
        lambda x: isinstance(x, str)
    ).values.all(), "Not all elements in the DataFrame are strings!"


from abc import ABC, abstractmethod
from typing import Optional, Any

from dsat.services.llm import LLMService

class Operator(ABC):
    """
    Abstract base class for a self-contained capability.

    Operators are the "verbs" of the agent framework, representing discrete
    actions like generating code, executing it, or reviewing results.
    """
    def __init__(self, llm_service: Optional[LLMService] = None, name: Optional[str] = None):
        self.name = name or self.__class__.__name__
        self.llm_service = llm_service

    @abstractmethod
    async def __call__(self, *args, **kwargs) -> Any:
        """
        Executes the operator's logic. All operator calls are asynchronous.
        """
        raise NotImplementedError


from pydantic import BaseModel, Field

# 1. Pydantic model for the optimizer's structured output
class GraphOptimize(BaseModel):
    modification: str = Field(description="A brief, one-sentence summary of the change made to the workflow code.")
    graph: str = Field(description="The complete, runnable Python code for the new 'Workflow' class.")

# 2. Prompt templates
# --- FIX START: MAKE PROMPT MORE GENERIC AND STRATEGY-FOCUSED ---
WORKFLOW_OPTIMIZE_PROMPT = """
You are an expert AI workflow engineer. Your task is to iteratively optimize a Python-based AI workflow to improve its problem-solving score.

You will be given the code of a parent workflow, its performance score, and a history of modifications.
Your goal is to propose a single, small, logical modification to the workflow code. The new code must be a complete and runnable Python class named `Workflow`.

RULES:
1.  **Focus on Logic**: Your modifications should improve the problem-solving STRATEGY. Examples: add a data cleaning step, try a different model, change a prompt, add a self-correction loop, etc.
2.  **Adhere to the Standard Interface**: The workflow you generate MUST correctly implement the `solve` method: `async def solve(self, description: str, io_instructions: str, data_dir: Path, output_path: Path)`.
3.  **Use the Arguments Correctly**:
    - `description`: Contains the task goal and a COMPLETE data analysis report.
    - `io_instructions`: Contains CRITICAL I/O requirements.
    Your code's logic MUST use both arguments effectively.
4.  **DO NOT Hardcode Filenames**: Your generated workflow code should extract the required output filename from the `output_path.name` attribute provided to the `solve` method, or ensure that any internal prompts correctly instruct the final code-generation step to do so.
5.  **Make only ONE logical change.** (e.g., add one new operator call, change a prompt, add a loop).
6.  **Analyze the experience log.** Avoid modifications that have failed in the past. Learn from successful ones.
7.  **Ensure the `graph` output is the complete, final Python code**, including all necessary imports and the full class definition.
8.  **Inherit from DSATWorkflow**: The generated class MUST be `class Workflow(DSATWorkflow):`.

Your response MUST be a JSON object that adheres to the provided schema. Do not include any other text, markdown, or explanations.
"""
# --- FIX END ---

WORKFLOW_INPUT_TEMPLATE = """
# PARENT WORKFLOW CONTEXT

## Experience Log
{experience}

## Parent Score
{score:.4f}

## Parent Code
```python
{graph_code}
```

## Available Operators
{operator_description}
"""

# 3. Helper functions to assemble the final prompt
def get_operator_description() -> str:
    """Returns a formatted string of available operators for the prompt."""
    return """
    You can call operators from the `self.operators` dictionary inside the `solve` method. For example: `final_answer = await self.operators['ScEnsemble'](...)`

    - `ScEnsemble`: Performs self-consistency voting to find the best solution from a list.
      - `__call__(self, solutions: List[str], problem: str) -> str`
    - `Review`: Critically reviews a solution and returns structured feedback.
      - `__call__(self, problem: str, solution: str) -> ReviewResponse(is_correct: bool, feedback: str)`
    - `Revise`: Revises a solution based on feedback.
      - `__call__(self, problem: str, solution: str, feedback: str) -> str`
    """

# --- FIX: Update function signature ---
def get_graph_optimize_prompt(experience: str, score: float, graph_code: str) -> str:
    """Assembles the full prompt for the optimizer LLM."""
    main_prompt = WORKFLOW_OPTIMIZE_PROMPT # Use the updated prompt directly
    
    inputs = WORKFLOW_INPUT_TEMPLATE.format(
        experience=experience,
        score=score,
        graph_code=graph_code,
        operator_description=get_operator_description()
    )
    return main_prompt + "\n" + inputs


#!/usr/bin/env python3
"""
DSLighting CLI - Command-line interface for DSLighting management.

Provides commands for package detection, configuration, and maintenance.
"""

import argparse
import logging
import sys
from pathlib import Path

# Setup logging
logging.basicConfig(
    level=logging.INFO,
    format='%(message)s'
)
logger = logging.getLogger(__name__)


def cmd_detect_packages(args):
    """Detect and save available packages to config."""
    from dslighting.utils.package_detector import PackageDetector

    config_path = Path(args.config) if args.config else None

    if config_path is None:
        # Find config.yaml
        for path in [Path.cwd(), Path.cwd().parent, Path.home() / ".dslighting"]:
            potential = path / "config.yaml"
            if potential.exists():
                config_path = potential
                break

    if config_path is None:
        config_path = Path.cwd() / "config.yaml"

    # Check if user wants all packages or only data science packages
    save_all = hasattr(args, 'all') and args.all

    if save_all:
        print(f"📦 Detecting ALL packages...")
        print(f"   Config: {config_path}")
        print(f"   Mode: Save all packages (including dependencies)")
    else:
        print(f"📦 Detecting Data Science & ML packages...")
        print(f"   Config: {config_path}")
        print(f"   Mode: Save only Data Science packages (recommended)")

    detector = PackageDetector()
    packages = detector.detect_packages()

    print(f"\n✓ Found {len(packages)} total packages in environment")

    # Show data science packages
    ds_packages = detector.get_data_science_packages()
    if ds_packages:
        print(f"\n📊 Data Science & ML Packages ({len(ds_packages)}):")
        for name, version in sorted(ds_packages.items()):
            print(f"   - {name} ({version})")

    if save_all:
        print(f"\n📦 Other Packages: {len(packages) - len(ds_packages)}")
        print(f"   (Use '--data-science-only' to save only DS packages)")

    # Save to config
    data_science_only = not save_all
    detector.save_to_config(config_path, packages, data_science_only=data_science_only)

    if save_all:
        print(f"\n✓ Saved all {len(packages)} packages to config: {config_path}")
    else:
        print(f"\n✓ Saved {len(ds_packages)} Data Science packages to config: {config_path}")

    # Show summary
    print(f"\n📝 Summary:")
    print(f"   Total packages in environment: {len(packages)}")
    print(f"   Data Science packages: {len(ds_packages)}")
    print(f"   Saved to config: {len(ds_packages) if not save_all else len(packages)}")
    print(f"   Config file: {config_path}")

    if not save_all:
        print(f"\n💡 Tip: Use 'dslighting detect-packages --all' to save all packages")
    else:
        print(f"\n💡 Tip: Use 'dslighting detect-packages --data-science-only' to save only DS packages")

    print(f"   Use 'agent = dslighting.Agent()' to automatically use this context")
    print(f"   Use 'agent = dslighting.Agent(include_package_context=False)' to disable")


def cmd_show_packages(args):
    """Show detected packages from config."""
    from dslighting.utils.package_detector import PackageDetector

    config_path = Path(args.config) if args.config else None

    if config_path is None:
        # Find config.yaml
        for path in [Path.cwd(), Path.cwd().parent, Path.home() / ".dslighting"]:
            potential = path / "config.yaml"
            if potential.exists():
                config_path = potential
                break

    if config_path is None or not config_path.exists():
        print(f"❌ Config file not found: {config_path}")
        print(f"   Run 'dslighting detect-packages' first")
        return 1

    print(f"📋 Loading packages from: {config_path}")

    detector = PackageDetector()
    packages = detector.load_from_config(config_path)

    if packages is None:
        print(f"❌ No packages found in config")
        print(f"   Run 'dslighting detect-packages' to detect packages")
        return 1

    print(f"\n✓ Found {len(packages)} packages in config")

    # Show formatted context
    print(f"\n{detector.format_as_context(packages)}")

    return 0


def cmd_validate_config(args):
    """Validate DSLighting configuration."""
    import yaml

    config_path = Path(args.config) if args.config else Path.cwd() / "config.yaml"

    if not config_path.exists():
        print(f"❌ Config file not found: {config_path}")
        return 1

    print(f"🔍 Validating: {config_path}")

    try:
        with open(config_path, 'r') as f:
            config = yaml.safe_load(f)

        # Check available_packages section
        if 'available_packages' in config:
            pkg_config = config['available_packages']

            print(f"\n✓ Package detection configured:")
            print(f"   Enabled: {pkg_config.get('enabled', True)}")
            print(f"   Packages: {len(pkg_config.get('packages', {}))}")

            if pkg_config.get('enabled', True):
                print(f"\n💡 Package context is ENABLED")
                print(f"   Agent will automatically know about available packages")
            else:
                print(f"\n⚠️  Package context is DISABLED")
                print(f"   Agent will not be aware of available packages")
                print(f"   Enable with: available_packages.enabled = true")
        else:
            print(f"\n⚠️  No package detection configured")
            print(f"   Run 'dslighting detect-packages' to set up")

        print(f"\n✓ Configuration is valid")

    except Exception as e:
        print(f"❌ Error: {e}")
        return 1

    return 0


def cmd_help(args):
    """Show DSLighting help and quick start guide."""
    print("=" * 70)
    print("DSLighting - Data Science Agent Framework")
    print("=" * 70)
    print()
    print("🚀 Quick Start:")
    print("-" * 70)
    print("""
from dotenv import load_dotenv
load_dotenv()

import dslighting

# Method 1: Use built-in dataset
data = dslighting.load_data("bike-sharing-demand")
agent = dslighting.Agent(workflow="aide")
result = agent.run(data)

# Method 2: Quick one-liner
result = dslighting.run_agent(task_id="bike-sharing-demand")
""")

    print("📋 Available Workflows:")
    print("-" * 70)
    print("""
  1. aide              - Adaptive Iteration & Debugging (Default)
  2. autokaggle        - Advanced competition solver
  3. data_interpreter  - Interactive data analysis
  4. automind          - Complex planning with knowledge base
  5. dsagent           - Long-term planning with logging
  6. deepanalyze       - Deep analysis with structured tags
""")

    print("💡 Useful Commands:")
    print("-" * 70)
    print("""
  dslighting workflows              - Show all workflows with details
  dslighting example <workflow>      - Show workflow example code
  dslighting quickstart             - Show detailed quick start guide
  dslighting detect-packages        - Detect available Python packages
  dslighting show-packages          - Show detected packages
""")

    print("📚 Documentation:")
    print("-" * 70)
    print("""
  Online:  https://luckyfan-cs.github.io/dslighting-web/
  GitHub:  https://github.com/usail-hkust/dslighting
  PyPI:    https://pypi.org/project/dslighting/
""")

    print("🔧 Python Help Functions:")
    print("-" * 70)
    print("""
  import dslighting
  dslighting.help()              - Show this help
  dslighting.list_workflows()    - List all workflows
  dslighting.show_example("aide") - Show workflow example
""")

    return 0


def cmd_workflows(args):
    """List all available workflows with details."""
    print("=" * 70)
    print("DSLighting Workflows")
    print("=" * 70)
    print()

    workflows = [
        {
            "name": "aide",
            "full_name": "AIDE (Adaptive Iteration & Debugging Enhancement)",
            "description": "Self-improving code with iterative debugging",
            "use_cases": ["Kaggle competitions (simple)", "Data analysis", "Quick prototyping"],
            "default_model": "gpt-4o",
            "parameters": {"max_iterations": 10, "temperature": 0.7},
            "unique_params": None
        },
        {
            "name": "autokaggle",
            "full_name": "AutoKaggle",
            "description": "Advanced competition solver with dynamic phase planning",
            "use_cases": ["Kaggle competitions (complex)", "High-stakes competitions", "Multi-phase problems"],
            "default_model": "gpt-4o",
            "parameters": {"temperature": 0.5},
            "unique_params": {
                "max_attempts_per_phase": "Max retries per phase (default: 5)",
                "success_threshold": "Score threshold 1-5 (default: 3.0)"
            }
        },
        {
            "name": "data_interpreter",
            "full_name": "DataInterpreter",
            "description": "Interactive data analysis and exploration",
            "use_cases": ["Data exploration", "Visualization", "Quick analysis"],
            "default_model": "gpt-4o-mini",
            "parameters": {"max_iterations": 5, "temperature": 0.7},
            "unique_params": None
        },
        {
            "name": "automind",
            "full_name": "AutoMind",
            "description": "Complex planning with knowledge base and experience replay",
            "use_cases": ["Complex tasks", "Multi-step problems", "Need historical context"],
            "default_model": "gpt-4o",
            "parameters": {"max_iterations": 10, "temperature": 0.5},
            "unique_params": {
                "case_dir": "Experience replay directory (e.g., ./experience_replay)"
            }
        },
        {
            "name": "dsagent",
            "full_name": "DS-Agent",
            "description": "Long-term planning with Plan-Execute-Log loop",
            "use_cases": ["Long-running tasks", "Need detailed logging", "Step-by-step refinement"],
            "default_model": "gpt-4o",
            "parameters": {"max_iterations": 15, "temperature": 0.6},
            "unique_params": {
                "case_dir": "Experience replay directory (e.g., ./experience_replay)"
            }
        },
        {
            "name": "deepanalyze",
            "full_name": "DeepAnalyze",
            "description": "Deep analysis with structured thinking tags",
            "use_cases": ["Deep data analysis", "Complex reasoning", "Structured outputs"],
            "default_model": "gpt-4o",
            "parameters": {"max_iterations": 10, "temperature": 0.8},
            "unique_params": None
        }
    ]

    for idx, wf in enumerate(workflows, 1):
        print(f"{idx}. {wf['name'].upper()}")
        print(f"   Full Name: {wf['full_name']}")
        print(f"   Description: {wf['description']}")
        print(f"   Use Cases: {', '.join(wf['use_cases'])}")
        print(f"   Default Model: {wf['default_model']}")

        if wf['unique_params']:
            print(f"   Unique Parameters:")
            for param, desc in wf['unique_params'].items():
                print(f"     - {param}: {desc}")
        else:
            print(f"   Unique Parameters: None (uses common params only)")

        print()

    print("💡 Use 'dslighting example <workflow>' to see example code")
    print("💡 Use 'dslighting help' for quick start guide")

    return 0


def cmd_example(args):
    """Show workflow example code."""
    workflow_name = args.workflow.lower()

    examples = {
        "aide": """
from dotenv import load_dotenv
load_dotenv()

import dslighting

data = dslighting.load_data("bike-sharing-demand")

agent = dslighting.Agent(
    workflow="aide",
    model="gpt-4o",
    temperature=0.7,
    max_iterations=10,
)

result = agent.run(data)

print(f"Score: {result.score}")
print(f"Cost: ${result.cost:.2f}")
""",
        "autokaggle": """
from dotenv import load_dotenv
load_dotenv()

import dslighting

data = dslighting.load_data("bike-sharing-demand")

agent = dslighting.Agent(
    workflow="autokaggle",
    model="gpt-4o",
    temperature=0.5,

    autokaggle={
        "max_attempts_per_phase": 5,
        "success_threshold": 3.5
    }
)

result = agent.run(data)

print(f"Score: {result.score}")
print(f"Duration: {result.duration:.1f}s")
print(f"Cost: ${result.cost:.2f}")
""",
        "data_interpreter": """
from dotenv import load_dotenv
load_dotenv()

import dslighting

data = dslighting.load_data("sales_data.csv")

agent = dslighting.Agent(
    workflow="data_interpreter",
    model="gpt-4o-mini",
    temperature=0.7,
    max_iterations=5,
)

result = agent.run(data, description="分析销售趋势")

print(f"Output: {result.output}")
print(f"Cost: ${result.cost:.2f}")
""",
        "automind": """
from dotenv import load_dotenv
load_dotenv()

import dslighting

data = dslighting.load_data("bike-sharing-demand")

agent = dslighting.Agent(
    workflow="automind",
    model="gpt-4o",
    temperature=0.5,
    max_iterations=10,

    automind={
        "case_dir": "./experience_replay"
    }
)

result = agent.run(data)

print(f"Score: {result.score}")
print(f"Output: {result.output}")
print(f"Cost: ${result.cost:.2f}")
""",
        "dsagent": """
from dotenv import load_dotenv
load_dotenv()

import dslighting

data = dslighting.load_data("bike-sharing-demand")

agent = dslighting.Agent(
    workflow="dsagent",
    model="gpt-4o",
    temperature=0.6,
    max_iterations=15,

    dsagent={
        "case_dir": "./experience_replay"
    }
)

result = agent.run(data)

print(f"Score: {result.score}")
print(f"Output: {result.output}")
print(f"Cost: ${result.cost:.2f}")
""",
        "deepanalyze": """
from dotenv import load_dotenv
load_dotenv()

import dslighting

data = dslighting.load_data("your_data.csv")

agent = dslighting.Agent(
    workflow="deepanalyze",
    model="gpt-4o",
    temperature=0.8,
    max_iterations=10,
)

result = agent.run(data, description="深度分析数据")

print(f"Output: {result.output}")
print(f"Cost: ${result.cost:.2f}")
"""
    }

    if workflow_name not in examples:
        print(f"❌ Unknown workflow: {args.workflow}")
        print(f"\nAvailable workflows: {', '.join(examples.keys())}")
        print(f"\nUse 'dslighting workflows' to see all workflows")
        return 1

    print("=" * 70)
    print(f"Example: {workflow_name.upper()}")
    print("=" * 70)
    print(examples[workflow_name])
    print("=" * 70)
    print(f"💡 Copy this code and run it!")
    print(f"💡 Use 'dslighting workflows' for more workflow details")

    return 0


def cmd_quickstart(args):
    """Show detailed quick start guide."""
    print("=" * 70)
    print("DSLighting Quick Start Guide")
    print("=" * 70)
    print()

    print("📦 Installation")
    print("-" * 70)
    print("""
pip install dslighting
""")

    print("🔑 Setup API Keys")
    print("-" * 70)
    print("""
# Create .env file in your project directory:
echo 'OPENAI_API_KEY=your_key_here' > .env
echo 'ANTHROPIC_API_KEY=your_key_here' >> .env
""")

    print("🚀 Your First Agent (3 Simple Steps)")
    print("-" * 70)
    print("""
# Step 1: Load environment variables
from dotenv import load_dotenv
load_dotenv()

# Step 2: Import DSLighting
import dslighting

# Step 3: Run your first agent
result = dslighting.run_agent(
    task_id="bike-sharing-demand",
    workflow="aide"
)

print(f"Success! Score: {result.score}")
""")

    print("📊 Using Your Own Data")
    print("-" * 70)
    print("""
import dslighting

# Load your data
data = dslighting.load_data("path/to/your/data.csv")

# Create agent
agent = dslighting.Agent(workflow="data_interpreter")

# Run with custom description
result = agent.run(data, description="分析销售数据并找出趋势")
""")

    print("🎯 Choosing the Right Workflow")
    print("-" * 70)
    print("""
  Task Type                     | Recommended Workflow
  ------------------------------|----------------------
  Quick analysis                | data_interpreter
  Simple competition            | aide
  Complex competition           | autokaggle
  Need historical knowledge     | automind
  Long-running task             | dsagent
  Deep analysis                 | deepanalyze
""")

    print("💡 Pro Tips")
    print("-" * 70)
    print("""
1. Start with 'aide' workflow for most tasks
2. Use 'gpt-4o-mini' for faster, cheaper results
3. Set max_iterations to control cost/time
4. Use keep_workspace=True to debug
5. Check results.score to evaluate performance
""")

    print("📚 Next Steps")
    print("-" * 70)
    print("""
  dslighting workflows           - See all workflows
  dslighting example aide        - See workflow examples
  dslighting help                - Show all commands

  https://luckyfan-cs.github.io/dslighting-web/  - Full documentation
""")

    return 0


def main():
    """Main CLI entry point."""
    parser = argparse.ArgumentParser(
        prog='dslighting',
        description='DSLighting - Data Science Agent CLI'
    )

    subparsers = parser.add_subparsers(dest='command', help='Available commands')

    # help command
    help_parser = subparsers.add_parser(
        'help',
        help='Show DSLighting help and quick start guide'
    )
    help_parser.set_defaults(func=cmd_help)

    # workflows command
    workflows_parser = subparsers.add_parser(
        'workflows',
        help='List all available workflows with details'
    )
    workflows_parser.set_defaults(func=cmd_workflows)

    # example command
    example_parser = subparsers.add_parser(
        'example',
        help='Show workflow example code'
    )
    example_parser.add_argument(
        'workflow',
        help='Workflow name (e.g., aide, autokaggle, data_interpreter, etc.)'
    )
    example_parser.set_defaults(func=cmd_example)

    # quickstart command
    quickstart_parser = subparsers.add_parser(
        'quickstart',
        help='Show detailed quick start guide'
    )
    quickstart_parser.set_defaults(func=cmd_quickstart)

    # detect-packages command
    detect_parser = subparsers.add_parser(
        'detect-packages',
        help='Detect and save available Python packages to config'
    )
    detect_parser.add_argument(
        '-c', '--config',
        help='Path to config.yaml file',
        default=None
    )
    detect_parser.add_argument(
        '--all',
        action='store_true',
        help='Save all packages (including dependencies). Default: save only Data Science packages'
    )
    detect_parser.add_argument(
        '--data-science-only',
        action='store_true',
        help='Save only Data Science & ML packages (default behavior)'
    )
    detect_parser.set_defaults(func=cmd_detect_packages)

    # show-packages command
    show_parser = subparsers.add_parser(
        'show-packages',
        help='Show detected packages from config'
    )
    show_parser.add_argument(
        '-c', '--config',
        help='Path to config.yaml file',
        default=None
    )
    show_parser.set_defaults(func=cmd_show_packages)

    # validate-config command
    validate_parser = subparsers.add_parser(
        'validate-config',
        help='Validate DSLighting configuration'
    )
    validate_parser.add_argument(
        '-c', '--config',
        help='Path to config.yaml file',
        default=None
    )
    validate_parser.set_defaults(func=cmd_validate_config)

    # Parse arguments
    args = parser.parse_args()

    if args.command is None:
        cmd_help(None)
        return 0

    # Execute command
    try:
        return args.func(args)
    except KeyboardInterrupt:
        print(f"\n\n⚠️  Interrupted")
        return 130
    except Exception as e:
        logger.error(f"Error: {e}")
        import traceback
        traceback.print_exc()
        return 1


if __name__ == '__main__':
    sys.exit(main())



import random
import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def _create_dataset_split(
    train_image_paths: list,
    test_image_paths: list,
    raw_dir: Path,
    public_dir: Path,
    private_dir: Path,
    steganography_algs: list,
):
    """
    Helper function to create a dataset split (e.g., train/test or train_val/test_val).

    This function populates the public and private directories with the respective
    training images, test images, and test set ground-truth labels.
    """
    # Prepare dirs
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)
    for subdir in ["Cover", "Test"] + steganography_algs:
        (public_dir / subdir).mkdir(exist_ok=True)

    # Copy over the train set for this split, giving them new ids
    for idx, fp in tqdm(
        enumerate(train_image_paths), total=len(train_image_paths), desc=f"Copying train images to {public_dir.name}"
    ):
        image_id = idx + 1  # 1-indexed
        shutil.copyfile(src=fp, dst=public_dir / "Cover" / f"{image_id:05d}.jpg")
        for alg in steganography_algs:
            shutil.copyfile(src=raw_dir / alg / fp.name, dst=public_dir / alg / f"{image_id:05d}.jpg")

    # Populate the test set for this split
    answers_rows = []
    random.seed(0)  # Reset seed for deterministic test set creation
    random.shuffle(test_image_paths)
    for idx, fp in tqdm(
        enumerate(test_image_paths), total=len(test_image_paths), desc=f"Copying test images to {public_dir.name}"
    ):
        image_id = idx + 1  # 1-indexed
        test_id = f"{image_id:04d}.jpg"
        dest = public_dir / "Test" / test_id

        # For the test set, we randomly select between the "Cover" (unedited image, negative class)
        # and one of the 3 steganography algorithms (positive class)
        # 1:1 ratio of positive:negative examples, and even distribution of steganography algorithms
        if random.choice([True, False]):
            # Negative class
            shutil.copyfile(
                src=fp,
                dst=dest,
            )
            answers_rows.append({"Id": test_id, "Label": 0})
        else:
            # Positive class
            alg = random.choice(steganography_algs)
            shutil.copyfile(src=raw_dir / alg / fp.name, dst=dest)
            answers_rows.append({"Id": test_id, "Label": 1})

    # Write answers to file
    answers_df = pd.DataFrame(answers_rows)
    answers_df.to_csv(private_dir / "test.csv", index=False)

    # Create sample submission
    sample_submission = answers_df.copy()
    sample_submission["Label"] = 0
    sample_submission.to_csv(public_dir / "sample_submission.csv", index=False)

    # Checks
    test_size = len(test_image_paths)
    assert "Id" in answers_df.columns, "Answers must have 'Id' column"
    assert "Label" in answers_df.columns, "Answers must have 'Label' column"
    assert "Id" in sample_submission.columns, "Sample submission must have 'Id' column"
    assert "Label" in sample_submission.columns, "Sample submission must have 'Label' column"
    assert (
        len(answers_df) == test_size
    ), f"Expected {test_size} test images, but got {len(answers_df)}"
    assert len(sample_submission) == len(
        answers_df
    ), f"Sample submission ({len(sample_submission)}) and answers ({len(answers_df)}) must have the same length"
    assert (
        len(list(public_dir.glob("Test/*.jpg"))) == test_size
    ), f"Expected {test_size} test images in {public_dir.name}/Test, but got {len(list(public_dir.glob('Test/*.jpg')))}"
    assert len(list(public_dir.glob("Cover/*.jpg"))) == len(
        train_image_paths
    ), f"Expected {len(train_image_paths)} train images in {public_dir.name}/Cover, but got {len(list(public_dir.glob('Cover/*.jpg')))}"
    for alg in steganography_algs:
        assert len(list(public_dir.glob(f"{alg}/*.jpg"))) == len(
            train_image_paths
        ), f"Expected {len(train_image_paths)} train images in {public_dir.name}/{alg}, but got {len(list(public_dir.glob(f'{alg}/*.jpg')))}"


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split (train_val/test_val) in parallel directories.
    """
    # List of all train image IDs
    cover_images_dir = raw / "Cover"
    cover_images = sorted(list(cover_images_dir.glob("*.jpg")))
    steganography_algs = ["JMiPOD", "JUNIWARD", "UERD"]
    test_size = 5000

    # --- Stage 1: Create the main competition train/test split ---
    # This split creates the final test set used for scoring.
    # The outputs in `public` and `private` are left untouched by subsequent steps.
    train_main, test_main = train_test_split(
        cover_images, test_size=test_size, random_state=42
    )

    _create_dataset_split(
        train_image_paths=train_main,
        test_image_paths=test_main,
        raw_dir=raw,
        public_dir=public,
        private_dir=private,
        steganography_algs=steganography_algs,
    )

    # --- Stage 2: Create the validation train/test split ---
    # This performs a second split on the main training data (`train_main`)
    # to create a new, smaller training set and a validation set.
    # Outputs are saved to `public_val` and `private_val` to avoid conflicts.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Split `train_main` again, using the same logic as the first split
    # to create a validation set of the same size as the main test set.
    train_val, test_val = train_test_split(
        train_main, test_size=test_size, random_state=42
    )

    # The new split is processed using the same helper function to ensure
    # identical directory structure, filenames, and creation logic.
    _create_dataset_split(
        train_image_paths=train_val,
        test_image_paths=test_val,
        raw_dir=raw,
        public_dir=public_val,
        private_dir=private_val,
        steganography_algs=steganography_algs,
    )


import pandas as pd

from mlebench.grade_helpers import InvalidSubmissionError


def pfbeta(labels, predictions, beta):
    """
    Adapted from https://www.kaggle.com/code/sohier/probabilistic-f-score
    adaptations: code formatting, this docstring.
    """
    y_true_count = 0
    ctp = 0
    cfp = 0

    for idx in range(len(labels)):
        prediction = min(max(predictions[idx], 0), 1)
        if labels[idx]:
            y_true_count += 1
            ctp += prediction
        else:
            cfp += prediction

    beta_squared = beta * beta
    c_precision = ctp / (ctp + cfp)
    c_recall = ctp / y_true_count
    if c_precision > 0 and c_recall > 0:
        result = (
            (1 + beta_squared) * (c_precision * c_recall) / (beta_squared * c_precision + c_recall)
        )
        return result
    else:
        return 0


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame) -> dict:
    id_col = "prediction_id"
    target_col = "cancer"

    assert id_col in answers.columns, f"{id_col} not in answers"
    assert target_col in answers.columns, f"{target_col} not in answers"

    if id_col not in submission.columns:
        raise InvalidSubmissionError(f"{id_col} not in submission")
    if target_col not in submission.columns:
        raise InvalidSubmissionError(f"{target_col} not in submission")
    if len(submission) != len(answers):
        raise InvalidSubmissionError("submission and answers have different lengths")

    submission = submission.sort_values(by=id_col)
    answers = answers.sort_values(by=id_col)

    if (submission[id_col].values != answers[id_col].values).any():
        raise InvalidSubmissionError(f"{id_col} not aligned")

    labels = answers[target_col].to_numpy()
    predictions = submission[target_col].to_numpy()

    return {"labels": labels, "predictions": predictions}


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    pf1_inputs = prepare_for_metric(submission, answers)
    # F1 score, so beta=1; https://www.kaggle.com/competitions/rsna-breast-cancer-detection/discussion/370123
    score = pfbeta(**pf1_inputs, beta=1)
    return score



import pandas as pd
import numpy as np
from sklearn.metrics import accuracy_score


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    """
    Grade the submission using accuracy metric.

    Args:
        submission: DataFrame with columns ['id', 'label']
        answers: DataFrame with columns ['id', 'label']

    Returns:
        Accuracy score (float between 0 and 1)
    """
    # Merge on id to ensure alignment
    merged = pd.merge(answers, submission, on='id', suffixes=('_true', '_pred'))

    # Calculate accuracy
    accuracy = accuracy_score(merged['label_true'], merged['label_pred'])

    return accuracy



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from mlebench.utils import read_csv


def verify_directory_sync(df: pd.DataFrame, expected_dir: Path, unexpected_dir: Path):
    """
    Checks that the dataframe contents match the directory structure.
    """
    for _, row in tqdm(
        df.iterrows(), desc=f"Verifying directory sync for {expected_dir.name}", total=len(df)
    ):
        case_day_path = expected_dir / row["case"] / f"{row['case']}_{row['day']}"
        assert (
            case_day_path.exists()
        ), f"Directory {case_day_path} does not exist but is listed in the dataframe."
        non_existent_path = unexpected_dir / row["case"] / f"{row['case']}_{row['day']}"
        assert (
            not non_existent_path.exists()
        ), f"Directory {non_existent_path} exists but is not listed in the dataframe."


def _create_split(
    input_df: pd.DataFrame,
    raw_images_dir: Path,
    output_public_path: Path,
    output_private_path: Path,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Helper function to perform a train/test split on a dataframe, move image files accordingly,
    and save the resulting CSVs and submission files.

    Args:
        input_df: The dataframe to be split.
        raw_images_dir: The source directory of all raw image data.
        output_public_path: The destination directory for public files (e.g., public/ or public_val/).
        output_private_path: The destination directory for private files (e.g., private/ or private_val/).

    Returns:
        A tuple containing the created training and testing dataframes.
    """
    # ----------------------- Splitting
    # Extract case and day from 'id'
    df_to_split = input_df.copy()
    df_to_split["case"] = df_to_split["id"].apply(lambda x: x.split("_")[0])
    df_to_split["day"] = df_to_split["id"].apply(lambda x: x.split("_")[1])
    df_to_split["slice"] = df_to_split["id"].apply(lambda x: x.split("_")[-1])

    # Split cases into train and test
    unique_cases = df_to_split["case"].unique()
    train_cases, test_cases = train_test_split(unique_cases, test_size=0.1, random_state=42)

    # Initially assign entire cases to train or test set
    df_to_split["set"] = df_to_split["case"].apply(lambda x: "test" if x in test_cases else "train")

    # Then mark some days from train to be test, to match competition test description
    days_df = df_to_split[df_to_split["set"] == "train"].groupby("case")["day"].apply(set).reset_index()
    for _, row in days_df.iterrows():
        # if theres more than 4 days, we will move any days past the 4th to the test set
        days = row["day"]
        if len(days) > 4:
            days = sorted(days, key=lambda x: int(x[len("day") :]))
            days_to_move = days[4:]
            # change their set to "test"
            df_to_split.loc[
                df_to_split["case"].eq(row["case"]) & df_to_split["day"].isin(days_to_move), "set"
            ] = "test"

    # ----------------------- Move the files to the correct new locations
    new_train_dir = output_public_path / "train"
    new_test_dir = output_public_path / "test"

    # Create new directories if they don't exist
    new_train_dir.mkdir(parents=True, exist_ok=True)
    new_test_dir.mkdir(parents=True, exist_ok=True)
    output_private_path.mkdir(parents=True, exist_ok=True)

    # Move directories based on the set assignment
    for case in tqdm(unique_cases, desc=f"Splitting by case for {output_public_path.name}"):
        original_path = raw_images_dir / case
        if case in train_cases:
            new_path = new_train_dir / case
        else:
            new_path = new_test_dir / case
        shutil.copytree(original_path, new_path, dirs_exist_ok=True)

    # Move specific days from public/train/ to public/test/ for marked case-days
    for _, row in tqdm(
        df_to_split.iterrows(),
        desc=f"Handling day-based splits for {output_public_path.name}",
        total=len(df_to_split),
    ):
        if row["set"] == "test":
            source_day_path = new_train_dir / row["case"] / f"{row['case']}_{row['day']}"
            target_day_path = new_test_dir / row["case"] / f"{row['case']}_{row['day']}"
            if source_day_path.exists():
                target_day_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.move(source_day_path.as_posix(), target_day_path.as_posix())

    # ------------------------ Saving splits
    new_train = df_to_split[df_to_split["set"] == "train"].copy()
    new_test = df_to_split[df_to_split["set"] == "test"].copy()
    # some asserts before we drop columns
    verify_directory_sync(new_train, expected_dir=new_train_dir, unexpected_dir=new_test_dir)
    verify_directory_sync(new_test, expected_dir=new_test_dir, unexpected_dir=new_train_dir)

    # get image height and image width for the test set, since this is needed for the metric
    for _, row in tqdm(
        new_test.iterrows(),
        desc=f"Getting image dimensions for {output_public_path.name} test set",
        total=len(new_test),
    ):
        case, day, day_slice = row["case"], row["day"], row["slice"]
        image_paths = list(
            (raw_images_dir / case / f"{case}_{day}" / "scans").glob(f"slice_{day_slice}_*.png")
        )
        assert len(image_paths) == 1, f"Expected 1 image, found {len(image_paths)}"
        image_path = image_paths[0]
        width, height = (int(length) for length in image_path.stem.split("_")[2:4])
        new_test.loc[row.name, "image_width"] = width
        new_test.loc[row.name, "image_height"] = height

    # dont need these anymore, and werent part of the original data
    new_train.drop(columns=["set", "case", "day", "slice"], inplace=True)
    new_test.drop(columns=["set", "case", "day", "slice"], inplace=True)

    # create sample submission
    sample_submission = new_test.copy()
    sample_submission["segmentation"] = "1 1 5 2"
    sample_submission.drop(columns=["image_height", "image_width"], inplace=True)
    sample_submission.rename(columns={"segmentation": "predicted"}, inplace=True)
    sample_submission.to_csv(output_public_path / "sample_submission.csv", index=False, na_rep="")

    # create private files
    new_test.rename(columns={"segmentation": "predicted"}, inplace=True)
    new_test.to_csv(output_private_path / "test.csv", index=False, na_rep="")

    # create public files
    new_train.to_csv(output_public_path / "train.csv", index=False, na_rep="")
    new_test_without_labels = new_test.drop(columns=["predicted", "image_width", "image_height"])
    new_test_without_labels.to_csv(output_public_path / "test.csv", index=False, na_rep="")

    # ------------------------ checks
    assert new_test_without_labels.shape[1] == 2, "Public test should have 2 columns."
    assert new_train.shape[1] == 3, "Public train should have 3 columns."
    assert len(new_train) + len(new_test) == len(
        input_df
    ), "Train and test should sum up to the original data."

    return new_train, new_test


def prepare(raw: Path, public: Path, private: Path):
    """
    Prepares the raw data by creating two sets of splits:
    1. A main train/test split for the final competition (`public`/`private`).
    2. A validation split from the main training data (`public_val`/`private_val`).
    """
    initial_train_df = read_csv(raw / "train.csv")
    raw_images_dir = raw / "train"

    # --- 1. Create the original train/test split ---
    # This generates the primary competition files in public/ and private/.
    # The output of this step will remain identical to the original script.
    print("--- Generating main train/test split for 'public' and 'private' directories ---")
    main_train_df, _ = _create_split(
        input_df=initial_train_df,
        raw_images_dir=raw_images_dir,
        output_public_path=public,
        output_private_path=private,
    )

    # --- 2. Create the validation split from the main training set ---
    # This takes the training data from the first split and splits it again,
    # creating a smaller training set and a validation set.
    # The outputs are saved in parallel directories to avoid conflicts.
    print("\n--- Generating validation split for 'public_val' and 'private_val' directories ---")
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    _create_split(
        input_df=main_train_df,  # Use the training set from the first split as input
        raw_images_dir=raw_images_dir,  # Image sources are the same
        output_public_path=public_val,
        output_private_path=private_val,
    )

    print("\nData preparation complete.")


"""
Utility modules for DSLighting.
"""

from dslighting.utils.defaults import (
    DEFAULT_WORKFLOW,
    DEFAULT_LLM_MODEL,
    DEFAULT_TEMPERATURE,
    DEFAULT_MAX_ITERATIONS,
    WORKFLOW_RECOMMENDATIONS,
)

__all__ = [
    "DEFAULT_WORKFLOW",
    "DEFAULT_LLM_MODEL",
    "DEFAULT_TEMPERATURE",
    "DEFAULT_MAX_ITERATIONS",
    "WORKFLOW_RECOMMENDATIONS",
]



import json
import random
import shutil
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.competitions.utils import get_logger, rle_encode

logger = get_logger(__name__)


def _create_answers_df(samples: list, raw_path: Path) -> pd.DataFrame:
    """Creates a DataFrame with ground truth answers for a given set of samples."""
    answers = []
    for sample in tqdm(samples, desc="Creating answers CSV"):
        record_id = sample["record_id"]
        mask = np.load(raw_path / "train" / record_id / "human_pixel_masks.npy")
        rle = rle_encode(mask)
        rle = " ".join([str(i) for i in rle]) if rle else "-"

        band = np.load(raw_path / "train" / record_id / "band_08.npy")
        height, width, _ = band.shape
        answers.append(
            {
                "record_id": record_id,
                "encoded_pixels": rle,
                "height": height,
                "width": width,
            }
        )
    return pd.DataFrame(answers)


def _copy_data_files(samples: list, raw_path: Path, dest_path: Path, include_masks: bool):
    """Copies data files (bands and optionally masks) for a given set of samples."""
    desc = f"Copying {'train' if include_masks else 'test'} files"
    for sample in tqdm(samples, desc=desc):
        record_id = sample["record_id"]
        (dest_path / record_id).mkdir(exist_ok=True, parents=True)
        # Copy bands
        for band_idx in range(8, 17):
            file_name = f"band_{band_idx:02}.npy"
            shutil.copyfile(
                src=raw_path / "train" / record_id / file_name,
                dst=dest_path / record_id / file_name,
            )
        if include_masks:
            # Copy human individual masks
            shutil.copyfile(
                src=raw_path / "train" / record_id / "human_individual_masks.npy",
                dst=dest_path / record_id / "human_individual_masks.npy",
            )
            # Copy human pixel masks
            shutil.copyfile(
                src=raw_path / "train" / record_id / "human_pixel_masks.npy",
                dst=dest_path / record_id / "human_pixel_masks.npy",
            )


def prepare(raw: Path, public: Path, private: Path):
    """
    We make train/test split from old train set, using same train/test proportion as the original
    competition. Concretely, the new split has 18673 train samples and 1856 test samples. We also
    copy over the validation set as-is.

    `sample_submission` is created with random predictions, either "1 3 10 5" or "-" (empty)

    A second split is performed on the train set to create a new validation set in parallel
    `public_val` and `private_val` directories.
    """
    DEV = False

    with open(raw / "train_metadata.json", "r") as f:
        train_metadata = json.load(f)
    train_metadata = train_metadata[:100] if DEV else train_metadata
    with open(raw / "validation_metadata.json", "r") as f:
        validation_metadata = json.load(f)

    # ==================================================================================
    # 1. Original Split: Create `new_train` and `new_test`
    # ==================================================================================
    if DEV:
        new_train, new_test = train_metadata[:90], train_metadata[90:]
    else:
        new_train, new_test = train_test_split(
            train_metadata, test_size=len(validation_metadata), random_state=0
        )
    logger.info(
        f"Created original split with {len(new_train)} train samples and {len(new_test)} test samples"
    )

    # ==================================================================================
    # 2. Process and Save Original Split to `public` and `private`
    # ==================================================================================
    public.mkdir(exist_ok=True)
    private.mkdir(exist_ok=True)

    # Copy train and test files
    _copy_data_files(new_train, raw, public / "train", include_masks=True)
    _copy_data_files(new_test, raw, public / "test", include_masks=False)

    # Create and save ground truth answers for the test set
    test_answers = _create_answers_df(new_test, raw)
    test_answers.to_csv(private / "answers.csv", index=False)

    # Save train metadata
    with open(public / "train_metadata.json", "w") as f:
        f.write(json.dumps(new_train))

    # Create and save a sample submission
    submission_df = test_answers.copy()
    random.seed(0)
    submission_df["encoded_pixels"] = [
        random.choice(["1 3 10 5", "-"]) for _ in range(len(submission_df))
    ]
    submission_df.to_csv(public / "sample_submission.csv", index=False)

    # Copy over existing validation data (this is unique to the original set)
    (raw / "validation").mkdir(exist_ok=True, parents=True)
    shutil.copytree(raw / "validation", public / "validation", dirs_exist_ok=True)
    shutil.copyfile(raw / "validation_metadata.json", public / "validation_metadata.json")

    # ==================================================================================
    # 3. New Validation Split: Split `new_train` into `train_val` and `test_val`
    # ==================================================================================
    train_val, test_val = train_test_split(
        new_train, test_size=len(new_test), random_state=0
    )
    logger.info(
        f"Created validation split with {len(train_val)} train_val samples and {len(test_val)} test_val samples"
    )

    # ==================================================================================
    # 4. Process and Save Validation Split to `public_val` and `private_val`
    # ==================================================================================
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    public_val.mkdir(exist_ok=True)
    private_val.mkdir(exist_ok=True)

    # Copy train_val and test_val files
    _copy_data_files(train_val, raw, public_val / "train", include_masks=True)
    _copy_data_files(test_val, raw, public_val / "test", include_masks=False)

    # Create and save ground truth answers for the test_val set
    test_val_answers = _create_answers_df(test_val, raw)
    # The filename must be "answers.csv" to mirror the private directory structure
    test_val_answers.to_csv(private_val / "answers.csv", index=False)

    # Save train_val metadata
    # The filename must be "train_metadata.json" to mirror the public directory structure
    with open(public_val / "train_metadata.json", "w") as f:
        f.write(json.dumps(train_val))

    # Create and save a sample submission for the validation set
    submission_val_df = test_val_answers.copy()
    random.seed(0)
    submission_val_df["encoded_pixels"] = [
        random.choice(["1 3 10 5", "-"]) for _ in range(len(submission_val_df))
    ]
    # The filename must be "sample_submission.csv" to mirror the public directory structure
    submission_val_df.to_csv(public_val / "sample_submission.csv", index=False)

    # ==================================================================================
    # 5. Sanity Checks
    # ==================================================================================
    logger.info("Performing sanity checks for original directories...")
    # Sanity checks for original directories
    assert (public / "train_metadata.json").exists(), "`train_metadata.json` doesn't exist!"
    assert (public / "sample_submission.csv").exists(), "`sample_submission.csv` doesn't exist!"
    assert (
        public / "validation_metadata.json"
    ).exists(), "`validation_metadata.json` doesn't exist!"
    assert (public / "train").exists(), "`train` directory doesn't exist!"
    assert (public / "test").exists(), "`test` directory doesn't exist!"
    assert (public / "validation").exists(), "`public` directory doesn't exist!"
    assert (private / "answers.csv").exists(), "`answers.csv` doesn't exist!"

    new_train_bands = list(img.stem for img in (public / "train").rglob("band*.npy"))
    assert (
        len(new_train_bands) == len(new_train) * 9
    ), f"Expected {len(new_train) * 9} bands in the train set, but got {len(new_train_bands)}!"
    new_test_bands = list(img.stem for img in (public / "test").rglob("band*.npy"))
    assert (
        len(new_test_bands) == len(new_test) * 9
    ), f"Expected {len(new_test) * 9} in the test set, but got {len(new_test_bands)}!"

    new_train_individual_masks = list(
        img.stem for img in (public / "train").rglob("human_individual_masks.npy")
    )
    assert len(new_train_individual_masks) == len(
        new_train
    ), f"Expected 1 human individual mask per sample in the train set, but got {len(new_train_individual_masks)}!"
    new_train_pixel_masks = list(
        img.stem for img in (public / "train").rglob("human_pixel_masks.npy")
    )
    assert len(new_train_pixel_masks) == len(
        new_train
    ), f"Expected 1 human pixel mask per sample in the train set, but got {len(new_train_pixel_masks)}!"

    logger.info("Performing sanity checks for validation directories...")
    # Sanity checks for new validation directories
    assert (public_val / "train_metadata.json").exists(), "`public_val/train_metadata.json` doesn't exist!"
    assert (public_val / "sample_submission.csv").exists(), "`public_val/sample_submission.csv` doesn't exist!"
    assert (public_val / "train").exists(), "`public_val/train` directory doesn't exist!"
    assert (public_val / "test").exists(), "`public_val/test` directory doesn't exist!"
    assert (private_val / "answers.csv").exists(), "`private_val/answers.csv` doesn't exist!"


import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def prepare(raw: Path, public: Path, private: Path):
    (public / "train").mkdir(exist_ok=True)
    (public / "test").mkdir(exist_ok=True)

    # Create train, test from train split. Ensure train, test come from different game plays
    old_train = pd.read_csv(raw / "train_labels.csv")
    unique_game_play = old_train["game_play"].unique()
    new_train_game_play, new_test_game_play = train_test_split(
        unique_game_play, test_size=0.1, random_state=0
    )

    new_train = old_train[old_train["game_play"].isin(new_train_game_play)]
    new_test = old_train[old_train["game_play"].isin(new_test_game_play)]
    assert set(new_train["contact_id"]).isdisjoint(
        set(new_test["contact_id"])
    ), "Train and test label share samples!"

    new_train.to_csv(public / "train_labels.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)

    # baseline helmets
    old_train_baseline_helmets = pd.read_csv(raw / "train_baseline_helmets.csv")
    new_train_baseline_helmets = old_train_baseline_helmets[
        old_train_baseline_helmets["game_play"].isin(new_train_game_play)
    ]
    new_test_baseline_helmets = old_train_baseline_helmets[
        old_train_baseline_helmets["game_play"].isin(new_test_game_play)
    ]

    new_train_baseline_helmets.to_csv(public / "train_baseline_helmets.csv", index=False)
    new_test_baseline_helmets.to_csv(public / "test_baseline_helmets.csv", index=False)

    # player tracking
    old_train_player_tracking = pd.read_csv(raw / "train_player_tracking.csv")
    new_train_player_trackings = old_train_player_tracking[
        old_train_player_tracking["game_play"].isin(new_train_game_play)
    ]
    new_test_player_trackings = old_train_player_tracking[
        old_train_player_tracking["game_play"].isin(new_test_game_play)
    ]

    new_train_player_trackings.to_csv(public / "train_player_tracking.csv", index=False)
    new_test_player_trackings.to_csv(public / "test_player_tracking.csv", index=False)

    # video metadata
    old_train_video_metadata = pd.read_csv(raw / "train_video_metadata.csv")
    new_train_video_metadata = old_train_video_metadata[
        old_train_video_metadata["game_play"].isin(new_train_game_play)
    ]
    new_test_video_metadata = old_train_video_metadata[
        old_train_video_metadata["game_play"].isin(new_test_game_play)
    ]

    new_train_video_metadata.to_csv(public / "train_video_metadata.csv", index=False)
    new_test_video_metadata.to_csv(public / "test_video_metadata.csv", index=False)

    # Copy over videos
    for game_play_type in ["All29", "Endzone", "Sideline"]:
        for game_play in new_train["game_play"].unique():
            shutil.copyfile(
                src=raw / "train" / f"{game_play}_{game_play_type}.mp4",
                dst=public / "train" / f"{game_play}_{game_play_type}.mp4",
            )

        for game_play in new_test["game_play"].unique():
            shutil.copyfile(
                src=raw / "train" / f"{game_play}_{game_play_type}.mp4",
                dst=public / "test" / f"{game_play}_{game_play_type}.mp4",
            )

    # Check integrity of the files copied
    num_train_videos_found = len(list(public.glob("train/*.mp4")))
    num_test_videos_found = len(list(public.glob("test/*.mp4")))
    num_expected_train_videos = (
        len(new_train["game_play"].unique()) * 3
    )  # *3 for All29, Endzone, Sideline
    num_expected_test_videos = len(new_test["game_play"].unique()) * 3

    assert (
        num_train_videos_found == num_expected_train_videos
    ), f"Expected {num_expected_train_videos} images, found {num_train_videos_found}"
    assert (
        num_test_videos_found == num_expected_test_videos
    ), f"Expected {num_expected_test_videos} images, found {num_test_videos_found}"

    # Create a sample submission file
    submission_df = pd.DataFrame(
        {
            "contact_id": new_test["contact_id"],
            "contact": 0,
        }
    )
    submission_df.to_csv(public / "sample_submission.csv", index=False)



# dsat/services/sandbox.py

import logging
import re
import subprocess
import sys
import uuid
import asyncio
import shutil
import time
from datetime import datetime
from contextlib import asynccontextmanager
from pathlib import Path
from typing import AsyncGenerator, List, Dict, Any, Optional
from multiprocessing import Process, Queue

import nbformat
from nbclient import NotebookClient
from nbclient.exceptions import CellExecutionError, CellTimeoutError, DeadKernelError

from dsat.common.typing import ExecutionResult
from dsat.services.workspace import WorkspaceService

logger = logging.getLogger(__name__)

NOTEBOOK_INIT_CODE = """
import warnings
import pandas as pd
import numpy as np
import os

# Optional: matplotlib (for plotting)
try:
    import matplotlib.pyplot as plt
    import matplotlib
    matplotlib.use('Agg')  # Use non-interactive backend
except Exception:
    pass

# Optional: seaborn (for better plots)
try:
    import seaborn as sns
    sns.set_theme(style="whitegrid")
except Exception:
    pass

warnings.filterwarnings('ignore')
print("DSAT Notebook environment initialized.")
"""

class NotebookExecutor:
    """
    Manages a persistent Jupyter kernel for cell-by-cell code execution.
    This class is intended to be used as an async context manager.
    """
    def __init__(self, workspace: WorkspaceService, timeout: int):
        self.workspace = workspace
        self.timeout = timeout
        self.nb = nbformat.v4.new_notebook()
        self.client = NotebookClient(
            self.nb, 
            timeout=self.timeout, 
            cwd=str(self.workspace.get_path("sandbox_workdir"))
        )
        self._initialized = False
        self._kernel_cm = None

    async def start(self):
        if self._initialized:
            return
        logger.info("Starting new Jupyter kernel for notebook execution...")
        self._kernel_cm = self.client.async_setup_kernel()
        await self._kernel_cm.__aenter__()
        logger.info("Kernel started. Initializing environment...")
        self._initialized = True
        sandbox_workdir_path = self.workspace.get_path("sandbox_workdir").resolve()
        full_init_code = (
            f"import os\n"
            f"os.chdir(r'{sandbox_workdir_path}')\n"
            f"print(f'CWD set to: {{os.getcwd()}}')\n"
            f"\n"
            f"{NOTEBOOK_INIT_CODE}"
        )
        init_result = await self.execute_cell(full_init_code)
        if not init_result.success:
            await self.stop()
            raise RuntimeError(f"Could not initialize notebook environment. Error: {init_result.stderr}")
        if self.nb.cells:
            self.nb.cells.pop(0)
        logger.info(f"Notebook environment ready. CWD is now {sandbox_workdir_path}")

    async def stop(self):
        logger.info("Shutting down Jupyter kernel.")
        if self._kernel_cm is not None:
            try:
                await self._kernel_cm.__aexit__(None, None, None)
            finally:
                self._kernel_cm = None
                self._initialized = False

    def _parse_cell_outputs(self, outputs: list) -> ExecutionResult:
        stdout_lines, stderr_lines, artifacts = [], [], []
        success = True
        exc_type = None
        for out in outputs:
            if out.output_type == 'stream':
                if out.name == 'stdout': stdout_lines.append(out.text)
                else: stderr_lines.append(out.text)
            elif out.output_type in ['execute_result', 'display_data']:
                if 'text/plain' in out.data: stdout_lines.append(out.data['text/plain'])
                if 'image/png' in out.data: artifacts.append("[Image data generated in notebook]")
            elif out.output_type == 'error':
                success = False
                exc_type = out.ename
                ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
                traceback_text = ansi_escape.sub('', "\n".join(out.traceback))
                stderr_lines.append(f"ERROR: {out.ename}\n{traceback_text}")
        return ExecutionResult(
            success=success, stdout="\n".join(stdout_lines).strip(),
            stderr="\n".join(stderr_lines).strip(), artifacts=artifacts, exc_type=exc_type
        )

    async def execute_cell(self, code: str) -> ExecutionResult:
        if not self._initialized:
            raise RuntimeError("NotebookExecutor not started. Use 'async with' context manager.")
        cell_index = len(self.nb.cells)
        self.nb.cells.append(nbformat.v4.new_code_cell(source=code))
        try:
            await self.client.async_execute_cell(self.nb.cells[cell_index], cell_index)
            result = self._parse_cell_outputs(self.nb.cells[cell_index].outputs)
        except CellExecutionError:
            result = self._parse_cell_outputs(self.nb.cells[cell_index].outputs)
        except (CellTimeoutError, DeadKernelError) as e:
            result = ExecutionResult(success=False, stderr=str(e), exc_type=e.__class__.__name__)
        
        notebook_path = self.workspace.get_path("artifacts") / "session.ipynb"
        with open(notebook_path, "w", encoding='utf-8') as f:
            nbformat.write(self.nb, f)
        return result

# ==============================================================================
# ==                    NEW: WORKER FUNCTION FOR SUBPROCESS                   ==
# ==============================================================================

def notebook_worker(
    task_queue: Queue,
    result_queue: Queue,
    run_dir: str,  # Changed: Pass run_dir path string instead of WorkspaceService object
    timeout: int
):
    """
    This function runs in a separate process. It creates an asyncio event loop,
    manages a NotebookExecutor, and processes tasks from the queue.

    IMPORTANT: On macOS (and any system using 'spawn' for multiprocessing),
    we must pass only picklable objects (like strings) to the worker process.
    We reconstruct WorkspaceService inside the worker from the run_dir path.
    """
    # Reconstruct WorkspaceService inside the worker process
    # We pass the parent directory as base_dir and run_dir name as run_name
    from pathlib import Path
    run_dir_path = Path(run_dir)
    base_dir = str(run_dir_path.parent)
    run_name = run_dir_path.name

    workspace = WorkspaceService(run_name, base_dir)

    async def main_loop():
        # The executor's lifecycle is tied to this async function
        executor = NotebookExecutor(workspace, timeout)
        await executor.start()
        
        while True:
            # Get code from the main process. A `None` object is the shutdown signal.
            code = task_queue.get()
            if code is None:
                break
            
            # Execute the code and put the result back
            result = await executor.execute_cell(code)
            result_queue.put(result)
            
        await executor.stop()

    try:
        asyncio.run(main_loop())
    except Exception as e:
        # If something catastrophic happens, report it back
        logger.error(f"Notebook worker process failed: {e}", exc_info=True)
        result_queue.put(ExecutionResult(
            success=False, stderr=f"Worker process failed: {e}", exc_type=e.__class__.__name__
        ))


class ProcessIsolatedNotebookExecutor:
    """
    Manages the lifecycle of a notebook worker process, providing a clean
    interface to the main application and ensuring robust cleanup.
    """
    def __init__(self, workspace: WorkspaceService, timeout: int):
        self.task_queue = Queue()
        self.result_queue = Queue()
        # Pass workspace path as string instead of WorkspaceService object
        # This ensures compatibility with macOS 'spawn' multiprocessing mode
        workspace_path = str(workspace.run_dir)
        self.worker_process = Process(
            target=notebook_worker,
            args=(self.task_queue, self.result_queue, workspace_path, timeout),
            daemon=True  # Set as daemon process to prevent main process from hanging
        )

    def start(self):
        """Starts the worker process."""
        logger.info("Starting process-isolated notebook worker...")
        self.worker_process.start()
        
    def stop(self, timeout: int = 10):
        """
        Stops the worker process gracefully, with a forceful termination fallback.
        Note: Since this is a daemon process, it will be automatically terminated
        when the parent process exits, providing an additional safety mechanism.
        """
        if not self.worker_process.is_alive():
            return
            
        logger.info("Stopping process-isolated notebook worker...")
        try:
            # Send shutdown signal
            self.task_queue.put(None)
            # Wait for graceful shutdown
            self.worker_process.join(timeout=timeout)
            
            # If still alive, it's hung. Terminate it.
            if self.worker_process.is_alive():
                logger.warning(
                    f"Notebook worker did not exit gracefully within {timeout}s. Terminating."
                )
                self.worker_process.terminate()
                self.worker_process.join() # Wait for termination to complete
        finally:
            self.task_queue.close()
            self.result_queue.close()
        logger.info("Notebook worker stopped.")

    async def execute_cell(self, code: str) -> ExecutionResult:
        """
        Sends code to the worker process and waits for the result.
        This is async to fit into the main application's event loop, but the
        underlying queue.get() is blocking. We use run_in_executor to avoid
        blocking the main event loop.
        """
        if not self.worker_process.is_alive():
            raise RuntimeError("Worker process is not running.")
            
        loop = asyncio.get_running_loop()
        
        # Send the task asynchronously
        await loop.run_in_executor(None, self.task_queue.put, code)
        
        # Wait for the result asynchronously
        result = await loop.run_in_executor(None, self.result_queue.get)
        
        return result


# ==============================================================================
# ==                   MODIFIED: SANDBOX SERVICE INTEGRATION                  ==
# ==============================================================================
class SandboxService:
    """
    Provides unified access to isolated script and notebook code execution environments.

    Args:
        workspace: Workspace service for managing files
        timeout: Default timeout for script execution (seconds)
        auto_matplotlib: Automatically inject matplotlib backend (default: False).
                        Set to True for Web UI environments that need visualization.
                        Set to False for standalone package usage.
    """
    def __init__(
        self,
        workspace: WorkspaceService,
        timeout: int = 600,
        auto_matplotlib: bool = False
    ):
        self.workspace = workspace
        self.timeout = timeout
        self.auto_matplotlib = auto_matplotlib  # Store matplotlib injection preference
        self.execution_history: List[Dict[str, Any]] = []

    def run_script(self, script_code: str, timeout: Optional[int] = None) -> ExecutionResult:
        """
        Runs a Python script within the sandbox workspace.

        Args:
            script_code: Python code to execute
            timeout: Optional timeout override (uses self.timeout if None)

        Returns:
            ExecutionResult with stdout, stderr, success status, etc.
        """
        # Optionally inject matplotlib non-interactive backend
        # This is only done if auto_matplotlib=True (used by Web UI for visualization)
        if self.auto_matplotlib:
            # Force non-interactive backend for matplotlib to prevent blocking plt.show()
            fixed_code = "import matplotlib\nmatplotlib.use('Agg')\n" + script_code
            logger.debug("Auto-injected matplotlib non-interactive backend")
        else:
            # Use code as-is without modification (default for DSLighting package)
            fixed_code = script_code

        script_name = f"_sandbox_script_{uuid.uuid4().hex}.py"
        script_path = self.workspace.run_dir / script_name
        execution_id = uuid.uuid4().hex
        started_at = datetime.utcnow()
        perf_start = time.perf_counter()
        execution_result: ExecutionResult = ExecutionResult(success=False, stdout="", stderr="", exc_type=None)

        try:
            script_path.write_text(fixed_code, encoding="utf-8")
            logger.info(f"Executing script '{script_name}' in sandbox (timeout: {self.timeout}s)...")
            completed_process = subprocess.run(
                [sys.executable, str(script_path)],
                capture_output=True, text=True, timeout=self.timeout,
                cwd=self.workspace.get_path("sandbox_workdir"),
                encoding='utf-8', errors='replace'
            )
            success = completed_process.returncode == 0
            exc_type = None
            if not success:
                stderr_lines = completed_process.stderr.strip().split('\n')
                if stderr_lines:
                    match = re.search(r"^(\w+(?:Error|Exception)):", stderr_lines[-1])
                    if match: exc_type = match.group(1)
            status = "succeeded" if success else f"failed (exit code {completed_process.returncode})"
            logger.info(f"Script execution finished: {status}.")
            if not success:
                logger.error(f"=== SCRIPT EXECUTION FAILED ===")
                logger.error(f"Exit Code: {completed_process.returncode}")
                logger.error(f"Exception Type: {exc_type}")
                if completed_process.stdout: logger.error(f"STDOUT:\n{completed_process.stdout}")
                if completed_process.stderr: logger.error(f"STDERR:\n{completed_process.stderr}")
                logger.error(f"=== END ERROR LOG ===")
            execution_result = ExecutionResult(
                success=success, stdout=completed_process.stdout,
                stderr=completed_process.stderr, exc_type=exc_type
            )
        except subprocess.TimeoutExpired as e:
            logger.warning("Script execution timed out. Process was terminated.")
            execution_result = ExecutionResult(
                success=False, stdout=e.stdout or "",
                stderr=e.stderr or f"TimeoutError: Execution exceeded {self.timeout} seconds.",
                exc_type="TimeoutError"
            )
        except Exception as e:
            logger.error(f"An unexpected error occurred during sandbox setup: {e}", exc_info=True)
            execution_result = ExecutionResult(success=False, stderr=str(e), exc_type=e.__class__.__name__)
        finally:
            ended_at = datetime.utcnow()
            duration = round(time.perf_counter() - perf_start, 4)

            copied_script_path = None
            scripts_dir = self.workspace.get_path("artifacts") / "sandbox_scripts"
            scripts_dir.mkdir(parents=True, exist_ok=True)
            if script_path.exists():
                try:
                    copied_script_path = scripts_dir / script_name
                    shutil.copy2(script_path, copied_script_path)
                except Exception as copy_error:
                    logger.error(f"Failed to copy sandbox script '{script_name}' to artifacts: {copy_error}", exc_info=True)

            execution_metadata = {
                "execution_id": execution_id,
                "script_filename": script_name,
                "original_script_path": str(script_path) if script_path.exists() else None,
                "copied_script_path": str(copied_script_path) if copied_script_path else None,
                "sandbox_cwd": str(self.workspace.get_path("sandbox_workdir")),
                "started_at_utc": started_at.isoformat() + "Z",
                "ended_at_utc": ended_at.isoformat() + "Z",
                "duration_seconds": duration,
            }
            execution_result.metadata = execution_metadata

            history_entry = {
                **execution_metadata,
                "success": execution_result.success,
                "exc_type": execution_result.exc_type,
                "stdout": execution_result.stdout,
                "stderr": execution_result.stderr,
                "code": fixed_code,
            }
            self.execution_history.append(history_entry)
        return execution_result

    @asynccontextmanager
    async def notebook_executor(self) -> AsyncGenerator[ProcessIsolatedNotebookExecutor, None]:
        """
        Provides a process-isolated notebook executor as an asynchronous context
        manager to ensure proper startup and cleanup of the worker process.
        """
        executor = ProcessIsolatedNotebookExecutor(self.workspace, self.timeout)
        executor.start()
        try:
            yield executor
        finally:
            executor.stop()

    def get_execution_history(self) -> List[Dict[str, Any]]:
        """Return a copy of sandbox execution history for telemetry persistence."""
        return list(self.execution_history)



import pandas as pd
from sklearn.metrics import root_mean_squared_error

from mlebench.grade_helpers import InvalidSubmissionError


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame) -> dict:
    id_col = "Id"
    target_col = "Pawpularity"

    assert id_col in answers.columns, f"Missing {id_col} column in answers"
    assert target_col in answers.columns, f"Missing {target_col} column in answers"
    assert (
        answers[target_col].between(1, 100).all()
    ), f"Pawpularity in answers should be between 1 and 100"

    if id_col not in submission.columns:
        raise InvalidSubmissionError(f"Missing {id_col} column in submission")

    if target_col not in submission.columns:
        raise InvalidSubmissionError(f"Missing {target_col} column in submission")

    if not submission[target_col].between(1, 100).all():
        raise InvalidSubmissionError(f"Pawpularity in submission should be between 1 and 100")

    submission = submission.sort_values(id_col)
    answers = answers.sort_values(id_col)

    if (submission[id_col].values != answers[id_col].values).any():
        raise InvalidSubmissionError(f"IDs in submission do not match IDs in answers")

    return {
        "y_true": answers[target_col].to_numpy(),
        "y_pred": submission[target_col].to_numpy(),
    }


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    rmse_input = prepare_for_metric(submission, answers)
    score = root_mean_squared_error(**rmse_input)
    return score



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv

from .classes import CLASSES


def _create_split_files(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    all_annotations_df: pd.DataFrame,
    public_dir: Path,
    private_dir: Path,
    raw_image_dir: Path,
):
    """
    Helper function to populate public and private directories for a given data split.

    This function handles directory creation, image copying, and CSV file generation,
    ensuring a consistent output structure.
    """
    # Ensure output directories exist
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)
    (public_dir / "train").mkdir(exist_ok=True)
    (public_dir / "test").mkdir(exist_ok=True)

    # Filter annotations to only include those for the current training set
    train_uids = train_df["StudyInstanceUID"]
    is_in_train = all_annotations_df["StudyInstanceUID"].isin(train_uids)
    split_train_annotations = all_annotations_df[is_in_train]

    # Copy image files for the current train and test sets
    for file_id in train_df["StudyInstanceUID"]:
        shutil.copyfile(
            src=raw_image_dir / f"{file_id}.jpg",
            dst=public_dir / "train" / f"{file_id}.jpg",
        )

    for file_id in test_df["StudyInstanceUID"]:
        shutil.copyfile(
            src=raw_image_dir / f"{file_id}.jpg",
            dst=public_dir / "test" / f"{file_id}.jpg",
        )

    # Assert that the correct number of images were copied
    assert len(list(public_dir.glob("train/*.jpg"))) == len(
        train_df
    ), f"Expected {len(train_df)} files in {public_dir}/train, got {len(list(public_dir.glob('train/*.jpg')))}"
    assert len(list(public_dir.glob("test/*.jpg"))) == len(
        test_df
    ), f"Expected {len(test_df)} files in {public_dir}/test, got {len(list(public_dir.glob('test/*.jpg')))}"

    # Create a sample submission file for the current test set
    submission_df = test_df[["StudyInstanceUID"] + CLASSES].copy()
    submission_df[CLASSES] = 0

    # Save all required CSV files with the required standard filenames
    train_df.to_csv(public_dir / "train.csv", index=False)
    split_train_annotations.to_csv(public_dir / "train_annotations.csv", index=False)
    submission_df.to_csv(public_dir / "sample_submission.csv", index=False)
    test_df.to_csv(private_dir / "test.csv", index=False)


def prepare(raw: Path, public: Path, private: Path):
    # Load raw data once
    old_train = read_csv(raw / "train.csv")
    old_train_annotations = read_csv(raw / "train_annotations.csv")
    raw_image_dir = raw / "train"

    # --- 1. Original Split: Create main train and test sets ---
    # This split creates the primary train/test sets for the competition.
    # The outputs in `public` and `private` must remain identical to the original script.
    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)

    # Use the helper to create the original public/private directory contents
    _create_split_files(
        train_df=new_train,
        test_df=new_test,
        all_annotations_df=old_train_annotations,
        public_dir=public,
        private_dir=private,
        raw_image_dir=raw_image_dir,
    )

    # --- 2. New Validation Split: Create a validation set from the main train set ---
    # Define paths for the new validation split directories.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Split the main training set (`new_train`) again to create a smaller
    # training set and a validation set.
    # The validation set (`test_val`) size should be ~10% of the *original* data.
    # new_train size = 0.9 * total. test_val_size = (0.1 * total) / (0.9 * total) = 1/9.
    test_val_size = 1 / 9
    train_val, test_val = train_test_split(
        new_train, test_size=test_val_size, random_state=0
    )

    # Use the same helper to create the new validation directories (`public_val`, `private_val`)
    # The helper ensures filenames and structure are identical to the original outputs.
    _create_split_files(
        train_df=train_val,
        test_df=test_val,
        all_annotations_df=old_train_annotations,
        public_dir=public_val,
        private_dir=private_val,
        raw_image_dir=raw_image_dir,
    )


"""
DSLighting Agent - Simplified API for data science automation.

This module provides the main Agent class that wraps the complexity of
DSAT framework while providing full control when needed.
"""

import asyncio
import logging
import time
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from dsat.config import DSATConfig
from dsat.models.task import TaskDefinition, TaskType
from dsat.runner import DSATRunner

from dslighting.core.config_builder import ConfigBuilder
from dslighting.core.data_loader import DataLoader, LoadedData

logger = logging.getLogger(__name__)


@dataclass
class AgentResult:
    """
    Result of running an Agent on a data science task.

    Attributes:
        success: Whether the task completed successfully
        output: Task output (predictions, answer, file path, etc.)
        score: Evaluation score (if available)
        cost: Total LLM cost in USD
        duration: Execution time in seconds
        artifacts_path: Path to generated artifacts
        workspace_path: Path to workspace directory
        error: Error message if failed
        metadata: Additional metadata
    """
    success: bool
    output: Any
    cost: float = 0.0
    duration: float = 0.0
    score: Optional[float] = None
    artifacts_path: Optional[Path] = None
    workspace_path: Optional[Path] = None
    error: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

    def __repr__(self) -> str:
        if self.success:
            return (
                f"AgentResult(success={self.success}, "
                f"output={self.output}, "
                f"score={self.score}, "
                f"cost=${self.cost:.4f}, "
                f"duration={self.duration:.1f}s)"
            )
        else:
            return (
                f"AgentResult(success={self.success}, "
                f"error={self.error}, "
                f"cost=${self.cost:.4f})"
            )


class Agent:
    """
    Simplified interface to DSLighting's data science automation capabilities.

    The Agent class provides a scikit-learn-like API that handles the complexity
    of workflow selection, configuration, and execution while allowing advanced
    users to access underlying components when needed.

    Examples:
        Simple usage:
            >>> import dslighting
            >>> agent = dslighting.Agent()
            >>> result = agent.run("data/my-competition")

        Advanced usage:
            >>> agent = dslighting.Agent(
            ...     workflow="autokaggle",
            ...     model="gpt-4o",
            ...     temperature=0.5,
            ...     max_iterations=10
            ... )
            >>> result = agent.run(data_path)

        Access underlying components:
            >>> config = agent.get_config()
            >>> runner = agent.get_runner()
    """

    def __init__(
        self,
        workflow: str = None,
        model: str = None,
        api_key: str = None,
        api_base: str = None,
        provider: str = None,
        temperature: float = None,
        max_iterations: int = None,
        num_drafts: int = None,
        workspace_dir: str = None,
        run_name: str = None,
        keep_workspace: bool = False,
        keep_workspace_on_failure: bool = True,
        verbose: bool = True,
        include_package_context: bool = True,
        **kwargs
    ):
        """
        Initialize DSLighting Agent.

        Args:
            workflow: Workflow name (aide, autokaggle, automind, etc.)
                      Defaults to "aide" or auto-detected from data
            model: LLM model name (e.g., "gpt-4o-mini", "deepseek-chat")
                   Defaults to LLM_MODEL env var or "gpt-4o-mini"
            api_key: API key for LLM service
                     Defaults to API_KEY env var
            api_base: API base URL
                      Defaults to API_BASE env var or OpenAI
            provider: LLM provider for LiteLLM (e.g., "siliconflow")
            temperature: LLM temperature (0.0-1.0)
            max_iterations: Maximum agent search iterations
            num_drafts: Number of drafts to generate
            workspace_dir: Custom workspace directory
            run_name: Name for this run
            keep_workspace: Keep workspace after completion (default: False)
            keep_workspace_on_failure: Keep workspace on failure (default: True)
            verbose: Enable verbose logging
            include_package_context: Include available package info in prompts (default: True)
            **kwargs: Additional parameters passed to DSATConfig
        """
        self.verbose = verbose
        self.logger = logger
        self.include_package_context = include_package_context

        # Build configuration
        self.config_builder = ConfigBuilder()
        self.config = self.config_builder.build_config(
            workflow=workflow,
            model=model,
            api_key=api_key,
            api_base=api_base,
            provider=provider,
            temperature=temperature,
            max_iterations=max_iterations,
            num_drafts=num_drafts,
            workspace_dir=workspace_dir,
            run_name=run_name,
            keep_workspace=keep_workspace,
            keep_workspace_on_failure=keep_workspace_on_failure,
            **kwargs
        )

        # Create runner (will be created on first use)
        self._runner: Optional[DSATRunner] = None

        # Track results
        self._results: List[AgentResult] = []

        # Initialize package detector for environment context
        self._package_detector = None
        self._package_context: Optional[str] = None

        if self.include_package_context:
            try:
                from dslighting.utils.package_detector import PackageDetector
                self._package_detector = PackageDetector()
                self.logger.info("Package context enabled: Agent will be aware of available packages")
            except Exception as e:
                self.logger.warning(f"Failed to initialize package detector: {e}")
                self.include_package_context = False

        self.logger.info(f"Agent initialized with workflow: '{self.config.workflow.name}'")

    def run(
        self,
        data: Union[str, Path, dict, pd.DataFrame, LoadedData] = None,
        task_id: str = None,
        data_dir: str = None,
        output_path: str = None,
        description: str = None,
        registry_dir: str = None,
        **kwargs
    ) -> AgentResult:
        """
        Run the agent on a data science task.

        This is the main entry point for executing data science tasks.
        It handles data loading, task creation, workflow execution, and
        result collection.

        Args:
            data: Optional data source (path, DataFrame, dict, or LoadedData).
                  If not provided, use task_id + data_dir pattern.
            task_id: Task/Competition identifier (e.g., "bike-sharing-demand").
                     Required when using MLE benchmark format.
            data_dir: Direct path to task data directory (must contain prepared/ folder).
                     Example: "/path/to/data/competitions/my-task"
                     If not provided, will use global config from setup().
            output_path: Custom output path for results
            description: Optional task description (overrides detected)
            registry_dir: Direct path to task registry directory (must contain config.yaml).
                          Example: "/path/to/registry/my-task"
                          If not provided, will use global config from setup().
            **kwargs: Additional task parameters

        Returns:
            AgentResult with output, metrics, and metadata

        Examples:
            >>> # Method 1: Using global setup (recommended for multiple tasks)
            >>> import dslighting
            >>> dslighting.setup(
            ...     data_parent_dir="/path/to/data/competitions",
            ...     registry_parent_dir="/path/to/registry"
            ... )
            >>>
            >>> agent = dslighting.Agent()
            >>> result = agent.run(task_id="bike-sharing-demand")

            >>> # Method 2: Using direct paths (explicit and clear)
            >>> result = agent.run(
            ...     task_id="bike-sharing-demand",
            ...     data_dir="/path/to/data/competitions/bike-sharing-demand",
            ...     registry_dir="/path/to/registry/bike-sharing-demand"
            ... )

            >>> # Method 3: Using built-in dataset
            >>> result = agent.run(task_id="bike-sharing-demand")

            >>> # Method 4: Using DataFrame
            >>> result = agent.run(df, description="Predict price")
        """
        # Start timing
        start_time = time.time()

        try:
            # ========== Path handling: Two clear modes ==========

            # Mode 1: Built-in dataset (only task_id provided, no data_dir)
            if task_id and data is None and data_dir is None:
                try:
                    import dslighting.datasets
                    # Convert "bike-sharing-demand" to "load_bike_sharing_demand"
                    load_func_name = f'load_{task_id.replace("-", "_")}'
                    load_func = getattr(dslighting.datasets, load_func_name, None)

                    if load_func:
                        self.logger.info(f"Using built-in dataset: {task_id}")
                        loaded_info = load_func()
                        # Load the built-in data
                        loader = DataLoader()
                        loaded_data = loader.load(loaded_info['data_dir'])
                        # Skip to execution (loaded_data is ready)
                        data = loaded_data
                        task_id = None  # Already extracted from loaded_data
                        data_dir = None
                except (AttributeError, FileNotFoundError):
                    # Not a built-in dataset, check global config
                    pass

            # Mode 2: Use global config if no data_dir provided
            if task_id and data is None and data_dir is None:
                from dslighting.core.global_config import get_global_config
                config = get_global_config()
                data_dir_from_config, registry_dir_from_config = config.get_task_paths(task_id)

                if data_dir_from_config is not None:
                    self.logger.info(f"Using global config for task: {task_id}")
                    self.logger.info(f"  Data directory: {data_dir_from_config}")
                    if registry_dir_from_config is not None:
                        self.logger.info(f"  Registry directory: {registry_dir_from_config}")
                        # Use registry from config if not explicitly provided
                        if registry_dir is None:
                            registry_dir = str(registry_dir_from_config)

                    # Load data from configured path
                    loader = DataLoader()
                    loaded_data = loader.load(data_dir_from_config, registry_dir=registry_dir)
                    data = loaded_data
                    task_id = None  # Already extracted from loaded_data
                    data_dir = None
                else:
                    raise ValueError(
                        f"No data directory provided and no global config found.\n"
                        f"Either:\n"
                        f"  1. Provide data_dir explicitly: agent.run(task_id='{task_id}', data_dir='/path/to/{task_id}')\n"
                        f"  2. Set up global config: dslighting.setup(data_parent_dir='/path/to/data/competitions')"
                    )

            # Mode 3: Direct path provided (data_dir is explicit)
            if task_id and data is None and data_dir is not None:
                data_dir_path = Path(data_dir).resolve()

                if not data_dir_path.exists():
                    raise FileNotFoundError(
                        f"Data directory not found: {data_dir_path}\n"
                        f"Please ensure the path exists and contains prepared/ folder."
                    )

                # Verify it's a task directory (has prepared/public)
                if not (data_dir_path / "prepared" / "public").exists():
                    raise ValueError(
                        f"Invalid data directory: {data_dir_path}\n"
                        f"Expected structure: {data_dir_path}/prepared/public/train.csv\n"
                        f"Please provide the direct task directory, not the parent directory."
                    )

                self.logger.info(f"Using direct task path: {data_dir_path}")
                if registry_dir:
                    self.logger.info(f"  Registry directory: {registry_dir}")

                # Load data with explicit paths
                loader = DataLoader()
                loaded_data = loader.load(data_dir_path, registry_dir=registry_dir)
                data = loaded_data
                task_id = None  # Already extracted from loaded_data
                data_dir = None

            # ========== Normal handling for remaining cases ==========
            if data is None:
                raise ValueError("Either 'data' or 'task_id' must be provided")

            # ========== Load data if needed ==========
            if not isinstance(data, LoadedData):
                self.logger.debug("Loading data...")
                loader = DataLoader()
                loaded_data = loader.load(data)
                self.logger.debug(f"Data loaded, task_id={loaded_data.task_id}")
            else:
                self.logger.debug("Data is already LoadedData")
                loaded_data = data
                self.logger.debug(f"LoadedData has task_id={loaded_data.task_id}")

            # Extract task_id from loaded_data if available
            if loaded_data.task_id:
                # Use task_id from loaded_data for benchmark initialization
                extracted_task_id = loaded_data.task_id
                self.logger.info(f"Detected task_id from data: {extracted_task_id}")
                self.logger.debug(f"Extracted task_id={extracted_task_id}, overriding task_id parameter")

                # Override task_id parameter for benchmark initialization
                task_id = extracted_task_id
                self.logger.debug(f"task_id is now set to: {task_id}")
            else:
                raise ValueError(
                    "Either 'task_id' or 'data' must be provided. "
                    "Example: agent.run(task_id='bike-sharing-demand', data_dir='data/competitions')"
                )

            # Get task information
            task_detection = loaded_data.task_detection

            # Determine workflow
            workflow = self._determine_workflow(task_detection)
            self.logger.info(f"Using workflow: {workflow}")

            # Create task definition
            task = self._create_task_definition(
                loaded_data=loaded_data,
                task_id=task_id,
                description=description,
                output_path=output_path,
                **kwargs
            )

            # Initialize benchmark for MLE tasks (for grading)
            self.logger.debug(f"Checking benchmark initialization: task_id={task_id}, task_type={loaded_data.get_task_type()}")
            self.logger.info(f"Checking benchmark initialization: task_id={task_id}, task_type={loaded_data.get_task_type()}")

            if task_id and loaded_data.get_task_type() == "kaggle":
                self.logger.debug(f"Condition met: task_id={task_id}, task_type=kaggle")
                self.logger.info(f"Initializing benchmark for task: {task_id}")
                try:
                    self.logger.debug("Attempting to import mlebench...")
                    from mlebench.grade import grade_csv
                    from mlebench.registry import Registry
                    self.logger.debug("mlebench imported successfully")

                    # Resolve data_dir - prioritize loaded_data.data_dir
                    if loaded_data.data_dir is not None:
                        # Extract parent directory from loaded_data.data_dir
                        # e.g., /path/to/competitions/bike-sharing-demand → /path/to/competitions
                        data_dir_path = loaded_data.data_dir.parent
                    elif data_dir is None:
                        data_dir_path = Path("data/competitions").expanduser().resolve()
                    else:
                        data_dir_path = Path(data_dir).expanduser().resolve()

                    self.logger.info(f"Initializing MLE-Bench grading for: {task_id}")
                    self.logger.info(f"  Data directory: {data_dir_path}")

                    # Try to use built-in registry first
                    import dslighting
                    built_in_registry_dir = Path(dslighting.__file__).parent / "registry"
                    self.logger.debug(f"Built-in registry dir: {built_in_registry_dir}, exists: {built_in_registry_dir.exists()}")

                    registry_dir = loaded_data.registry_dir
                    self.logger.debug(f"loaded_data.registry_dir: {registry_dir}")
                    if not registry_dir or not Path(registry_dir).exists():
                        self.logger.debug("Registry dir not valid, checking built-in...")
                        if built_in_registry_dir.exists():
                            registry_dir = built_in_registry_dir
                            self.logger.info(f"  Using built-in registry: {registry_dir}")
                            self.logger.debug(f"Using built-in registry: {registry_dir}")
                        else:
                            self.logger.warning(f"⚠️  Registry directory not found")
                            self.logger.warning(f"  Grading will be skipped.")
                            self.logger.debug("Built-in registry not found, skipping grading")
                    else:
                        self.logger.info(f"  Registry directory (from loaded_data): {registry_dir}")
                        self.logger.debug(f"Using loaded_data registry: {registry_dir}")

                    # Check if registry_dir exists
                    if not registry_dir or not Path(registry_dir).exists():
                        self.logger.warning(f"⚠️  Registry directory not available: {registry_dir}")
                        self.logger.warning(f"   Grading will be skipped.")
                        self.logger.debug("Registry not available, skipping grading")
                    else:
                        self.logger.debug("Registry available, proceeding with initialization")
                        # Initialize benchmark with registry (keep as Path objects)
                        registry_dir_path = Path(registry_dir) if not isinstance(registry_dir, Path) else registry_dir
                        registry_kwargs = {"data_dir": data_dir_path, "registry_dir": registry_dir_path}

                        try:
                            self.logger.debug(f"Creating Registry with kwargs: {registry_kwargs}")
                            custom_registry = Registry(**registry_kwargs)
                            self.logger.debug("Registry created successfully")

                            # Log registry configuration
                            self.logger.info(f"  Registry config:")
                            self.logger.info(f"    data_dir: {data_dir_path}")
                            self.logger.info(f"    registry_dir: {registry_dir}")

                            # Create simple wrapper class
                            class SimpleMLEBenchmark:
                                def __init__(self, registry_instance, logger, registry_dir, data_dir, task_id):
                                    self.registry = registry_instance
                                    self.problems = [{"competition_id": task_id}]
                                    self.logger = logger
                                    self.registry_dir = registry_dir
                                    self.data_dir = data_dir
                                    self.task_id = task_id

                                async def grade(self, submission_path):
                                    """Grade submission using custom registry."""
                                    try:
                                        # Read config directly from registry
                                        import yaml
                                        config_path = self.registry_dir / self.task_id / "config.yaml"

                                        if not config_path.exists():
                                            self.logger.warning(f"  Config not found: {config_path}")
                                            return 0.0

                                        with open(config_path) as f:
                                            config = yaml.safe_load(f)

                                        # Resolve paths relative to data_dir (from config.yaml)
                                        # self.data_dir is the parent directory (e.g., /path/to/competitions)
                                        # config["dataset"]["answers"] is relative path like "bike-sharing-demand/prepared/private/test_answer.csv
                                        answers_rel_path = config.get("dataset", {}).get("answers", "")
                                        answers_path = self.data_dir / answers_rel_path

                                        # **MANDATORY**: Check for prepared/public and prepared/private structure
                                        competition_dir = self.data_dir / self.task_id
                                        prepared_public_dir = competition_dir / "prepared" / "public"
                                        prepared_private_dir = competition_dir / "prepared" / "private"

                                        if not prepared_public_dir.exists():
                                            self.logger.error(f"  ❌ Required directory not found: {prepared_public_dir}")
                                            self.logger.error(f"  ❌ Tasks must have prepared/public/ directory structure")
                                            self.logger.error(f"  See: https://github.com/usail-hkust/dslighting for setup instructions")
                                            return 0.0

                                        if not prepared_private_dir.exists():
                                            self.logger.error(f"  ❌ Required directory not found: {prepared_private_dir}")
                                            self.logger.error(f"  ❌ Tasks must have prepared/private/ directory structure")
                                            self.logger.error(f"  See: https://github.com/usail-hkust/dslighting for setup instructions")
                                            return 0.0

                                        self.logger.info(f"  ✓ Required structure verified:")
                                        self.logger.info(f"    - prepared/public: {prepared_public_dir}")
                                        self.logger.info(f"    - prepared/private: {prepared_private_dir}")

                                        if not answers_path.exists():
                                            self.logger.warning(f"  Answers file not found: {answers_path}")
                                            self.logger.warning(f"  Looking for: {answers_path}")
                                            return 0.0

                                        self.logger.info(f"  ✓ Found answers file: {answers_path}")

                                        # Import the actual Competition class from mlebench
                                        from mlebench.registry import Competition
                                        from mlebench.grade_helpers import Grader

                                        # Load grader
                                        grader_config = config.get("grader", {})
                                        grader_name = grader_config.get("name", "rmsle")

                                        # Import grade function if specified
                                        grade_fn = None
                                        if "grade_fn" in grader_config:
                                            # Parse grade_fn format: mlebench.competitions.bike_sharing_demand.grade:grade
                                            fn_str = grader_config["grade_fn"]
                                            if ":" in fn_str:
                                                module_path, fn_name = fn_str.rsplit(":", 1)
                                                # Convert to file import if needed
                                                if not module_path.startswith("file:"):
                                                    fn_file = self.registry_dir / self.task_id / "grade.py"
                                                    if fn_file.exists():
                                                        fn_str = f"file:{fn_file}:{fn_name}"
                                                else:
                                                    # Try to import from mlebench
                                                    try:
                                                        import importlib
                                                        importlib.import_module(module_path)
                                                    except:
                                                        pass

                                        # Create a simple grader
                                        if grade_fn or fn_str:
                                            grader = Grader(
                                                name=grader_name,
                                                grade_fn=fn_str if fn_str else grade_fn,
                                            )
                                        else:
                                            # Default RMSLE grader
                                            grader = Grader(name="rmsle", grade_fn=None)

                                        # Resolve paths - use actual prepared directories (already verified above)
                                        raw_dir = competition_dir / "raw"
                                        checksums = competition_dir / "checksums.txt"
                                        leaderboard = competition_dir / "leaderboard.csv"
                                        # Use the actual prepared directories that we verified exist
                                        private_dir = prepared_private_dir
                                        public_dir = prepared_public_dir

                                        # Create placeholder prepare_fn
                                        def dummy_prepare_fn(a, b, c):
                                            return private_dir

                                        # Create actual Competition object with all required fields
                                        simple_comp = Competition(
                                            id=config["id"],
                                            name=config["name"],
                                            description=config.get("description", ""),
                                            grader=grader,
                                            answers=answers_path,
                                            gold_submission=answers_path,  # Use same as answers for grading
                                            sample_submission=public_dir / "sampleSubmission.csv",
                                            competition_type=config.get("competition_type", "standard"),
                                            prepare_fn=dummy_prepare_fn,
                                            raw_dir=raw_dir,
                                            private_dir=private_dir,
                                            public_dir=public_dir,
                                            checksums=checksums,
                                            leaderboard=leaderboard,
                                        )

                                        # Grade using mlebench's grade_csv
                                        report = grade_csv(
                                            submission_path,
                                            simple_comp,
                                        )
                                        # Return the score (float), not the entire report
                                        score = report.score if report.score is not None else 0.0
                                        self.logger.info(f"  Grading result: {score}")
                                        return score
                                    except Exception as e:
                                        self.logger.warning(f"  Grading failed: {e}")
                                        import traceback
                                        self.logger.warning(f"  Traceback: {traceback.format_exc()}")
                                        return 0.0

                            benchmark = SimpleMLEBenchmark(custom_registry, self.logger, registry_dir_path, data_dir_path, task_id)
                            runner = self.get_runner()
                            runner.benchmark = benchmark
                            self.logger.debug(f"✓ Benchmark set successfully for task: {task_id}")
                            self.logger.info(f"✓ Benchmark initialized for grading: {task_id}")
                        except Exception as e:
                            self.logger.debug(f"Benchmark initialization failed: {e}")
                            self.logger.warning(f"⚠️  Benchmark initialization failed: {e}")
                            self.logger.warning(f"   Grading will be skipped.")

                except ImportError as e:
                    self.logger.warning(f"MLE-Bench import failed: {e}")
                    self.logger.warning("Grading will be skipped.")
                except Exception as e:
                    self.logger.warning(f"Benchmark initialization failed: {e}")
                    self.logger.warning(f"Grading will be skipped: {e}")

            # Execute task (async wrapper)
            result = asyncio.run(self._execute_task(task, loaded_data))

            # Calculate duration
            duration = time.time() - start_time
            result.duration = duration

            # Store result
            self._results.append(result)

            if self.verbose:
                self._log_result(result)

            return result

        except Exception as e:
            self.logger.error(f"Task execution failed: {e}", exc_info=True)

            duration = time.time() - start_time

            return AgentResult(
                success=False,
                output=None,
                duration=duration,
                error=str(e),
                metadata={"exception_type": type(e).__name__}
            )

    def run_batch(
        self,
        data_list: List[Union[str, Path, dict, pd.DataFrame]],
        **kwargs
    ) -> List[AgentResult]:
        """
        Run the agent on multiple data sources sequentially.

        Args:
            data_list: List of data sources
            **kwargs: Additional parameters passed to run()

        Returns:
            List of AgentResult objects

        Examples:
            >>> results = agent.run_batch([
            ...     "data/titanic",
            ...     "data/house-prices",
            ...     "data/fraud-detection"
            ... ])
            >>> for r in results:
            ...     print(f"{r.task_id}: {r.score}")
        """
        results = []

        for i, data in enumerate(data_list):
            self.logger.info(f"Running batch task {i+1}/{len(data_list)}")

            result = self.run(data, **kwargs)
            results.append(result)

        return results

    def get_config(self) -> DSATConfig:
        """
        Get the underlying DSAT configuration.

        This allows advanced users to modify configuration directly.

        Returns:
            DSATConfig object

        Examples:
            >>> config = agent.get_config()
            >>> config.llm.temperature = 0.5
        """
        return self.config

    def get_runner(self) -> DSATRunner:
        """
        Get the underlying DSATRunner instance.

        This allows advanced users to interact with the runner directly.

        Returns:
            DSATRunner instance

        Examples:
            >>> runner = agent.get_runner()
            >>> eval_fn = runner.get_eval_function()
        """
        if self._runner is None:
            self._runner = DSATRunner(self.config)

        return self._runner

    def get_results(self) -> List[AgentResult]:
        """Get all results from this agent session."""
        return self._results.copy()

    def _determine_workflow(self, task_detection) -> str:
        """Determine which workflow to use."""
        # If user specified workflow, use it
        if self.config.workflow and self.config.workflow.name:
            return self.config.workflow.name

        # Otherwise, use recommended workflow from detection
        if task_detection and task_detection.recommended_workflow:
            return task_detection.recommended_workflow

        # Fallback to default
        return "aide"

    def _create_task_definition(
        self,
        loaded_data: LoadedData,
        task_id: str = None,
        description: str = None,
        output_path: str = None,
        **kwargs
    ) -> TaskDefinition:
        """Create TaskDefinition from LoadedData."""
        # Generate task ID if not provided
        if task_id is None:
            safe_name = str(uuid.uuid4())[:8]
            task_id = f"task_{safe_name}"

        # Get task type
        task_type = loaded_data.get_task_type()

        # Get description
        if description is None:
            description = loaded_data.get_description()

        # Add package context to description if enabled
        if self.include_package_context and self._package_detector:
            try:
                package_context = self._package_detector.format_as_context()
                # Prepend package context to description
                description = f"{package_context}\n\nTask Description:\n{description}"
                self.logger.info("Package context added to task description")
            except Exception as e:
                self.logger.warning(f"Failed to add package context: {e}")

        # Get I/O instructions
        io_instructions = loaded_data.get_io_instructions()

        # Build payload
        payload = kwargs.copy()
        payload["description"] = description
        payload["io_instructions"] = io_instructions

        # Add data directory based on task type
        if loaded_data.data_dir:
            data_dir = loaded_data.data_dir

            if task_type == "kaggle":
                # MLE/Kaggle format: needs public_data_dir and output_submission_path
                # Follow MLEBenchmark pattern: {data_dir}/prepared/public
                prepared_dir = data_dir / "prepared"
                public_dir = prepared_dir / "public"

                # Check if prepared/public exists (MLE format)
                if public_dir.exists():
                    payload["public_data_dir"] = str(public_dir.resolve())
                    self.logger.info(f"Using MLE prepared data: {public_dir.resolve()}")
                else:
                    # Fallback: use data_dir directly
                    payload["public_data_dir"] = str(data_dir.resolve())
                    self.logger.warning(
                        f"Prepared data not found at {public_dir}, using data_dir instead"
                    )

                # Set output path - use simple filename, will be saved in workspace/sandbox
                if output_path is None:
                    # Extract competition_id from data_dir path if possible
                    competition_id = data_dir.name
                    unique_id = str(uuid.uuid4())[:8]
                    output_filename = f"submission_{competition_id}_{unique_id}.csv"

                    # Use just the filename - DSAT will save it in workspace/sandbox
                    output_path = Path(output_filename)

                payload["output_submission_path"] = str(output_path)
                self.logger.info(f"Output submission file: {output_path}")
            else:
                # Other task types: use data_dir
                payload["data_dir"] = str(data_dir)

        # Add output path if specified (for non-kaggle tasks)
        if output_path and task_type != "kaggle":
            payload["output_submission_path"] = str(output_path)

        # Create TaskDefinition
        task = TaskDefinition(
            task_id=task_id,
            task_type=task_type,  # Pass string directly, Pydantic will validate
            payload=payload
        )

        return task

    def _get_workspace_dir(self) -> Path:
        """Get the workspace directory for this agent run."""
        # Try to get from config
        workspace_dir = None

        if hasattr(self, 'config') and hasattr(self.config, 'run'):
            run_config = self.config.run
            if hasattr(run_config, 'parameters') and run_config.parameters:
                workspace_dir = run_config.parameters.get('workspace_dir')

        # Fallback to default workspace directory
        if workspace_dir is None:
            from dslighting.utils.defaults import DEFAULT_WORKSPACE_DIR
            workspace_dir = DEFAULT_WORKSPACE_DIR

        # Ensure workspace exists
        workspace_path = Path(workspace_dir)
        workspace_path.mkdir(parents=True, exist_ok=True)

        return workspace_path

    def _get_default_benchmark_dir(self) -> Path:
        """
        Get the default benchmark registry directory.

        This is where task registration files (grade.py, description.md, etc.) are stored.
        Priority:
        1. Built-in registry in dslighting package (dslighting/registry/)
        2. Local benchmarks/mlebench/competitions/
        3. Config-provided benchmark_dir

        Returns:
            Path to benchmark registry directory
        """
        # Try to get from config
        benchmark_dir = None

        if hasattr(self, 'config') and hasattr(self.config, 'run'):
            run_config = self.config.run
            if hasattr(run_config, 'parameters') and run_config.parameters:
                benchmark_dir = run_config.parameters.get('benchmark_dir')

        # Fallback 1: Try built-in registry in dslighting package
        if benchmark_dir is None:
            try:
                import dslighting
                dslighting_path = Path(dslighting.__file__).parent
                built_in_registry = dslighting_path / "registry"

                if built_in_registry.exists():
                    self.logger.info(f"Using built-in registry: {built_in_registry}")
                    return built_in_registry.resolve()
            except Exception as e:
                self.logger.debug(f"Could not access built-in registry: {e}")

        # Fallback 2: Use relative path from current working directory
        if benchmark_dir is None:
            # Default: benchmarks/mlebench/competitions/
            benchmark_dir = "benchmarks/mlebench/competitions"

        benchmark_path = Path(benchmark_dir).resolve()

        self.logger.debug(f"Benchmark registry directory: {benchmark_path}")

        return benchmark_path

    async def _execute_task(
        self,
        task: TaskDefinition,
        loaded_data: LoadedData
    ) -> AgentResult:
        """Execute a single task using DSATRunner."""
        runner = self.get_runner()

        # Link data directory to workspace if available
        if loaded_data.data_dir and hasattr(runner, 'benchmark') and runner.benchmark:
            workspace_service = None
            # Try to get workspace service from workflow
            if hasattr(runner, 'factory'):
                try:
                    workflow = runner.factory.create_workflow(self.config, benchmark=runner.benchmark)
                    workspace_service = workflow.services.get("workspace")
                    if workspace_service:
                        workspace_service.link_data_to_workspace(loaded_data.data_dir)
                except Exception:
                    pass

        # Get evaluation function
        eval_fn = runner.get_eval_function()

        # Execute task
        output, cost, usage_summary = await eval_fn(task)

        # Check for errors
        if isinstance(output, str) and output.startswith("[ERROR]"):
            return AgentResult(
                success=False,
                output=output,
                cost=cost,
                error=output
            )

        # Extract score if available
        score = None
        if isinstance(output, (int, float)):
            score = float(output)
        elif isinstance(output, dict) and "score" in output:
            score = output["score"]

        # Get workspace path
        workspace_path = None
        if hasattr(runner, 'run_records') and runner.run_records:
            last_record = runner.run_records[-1]
            workspace_path = last_record.get("workspace_dir")

        return AgentResult(
            success=True,
            output=output,
            cost=cost,
            score=score,
            workspace_path=Path(workspace_path) if workspace_path else None,
            metadata={"usage": usage_summary}
        )

    def _log_result(self, result: AgentResult):
        """Log result summary."""
        if result.success:
            self.logger.info(
                f"✓ Task completed successfully | "
                f"Score: {result.score or 'N/A'} | "
                f"Cost: ${result.cost:.4f} | "
                f"Duration: {result.duration:.1f}s"
            )
        else:
            self.logger.error(
                f"✗ Task failed | "
                f"Error: {result.error} | "
                f"Cost: ${result.cost:.4f}"
            )

    def __repr__(self) -> str:
        return (
            f"Agent(workflow='{self.config.workflow.name}', "
            f"model='{self.config.llm.model}', "
            f"results={len(self._results)})"
        )



import pandas as pd
import re


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    """
    Grade the DABench submission.

    Args:
        submission: DataFrame with columns ['id', 'answer']
        answers: DataFrame with columns ['id', 'answer']

    Returns:
        float: 1.0 if exact match, 0.0 otherwise
    """
    try:
        # Both should have exactly one row
        if len(submission) != 1 or len(answers) != 1:
            return 0.0

        # Get the submission and answer strings
        submission_str = str(submission.iloc[0]['answer']).strip()
        answer_str = str(answers.iloc[0]['answer']).strip()

        # Parse all key-value pairs
        # Expected keys: ['outlier_countries']
        pattern = r'@(\w+)\[([^\]]+)\]'

        submission_dict = dict(re.findall(pattern, submission_str))
        answer_dict = dict(re.findall(pattern, answer_str))

        if not submission_dict or not answer_dict:
            print(f"Failed to parse: submission='{submission_str}', answer='{answer_str}'")
            return 0.0

        # Check if all keys match
        if set(submission_dict.keys()) != set(answer_dict.keys()):
            print(f"Key mismatch: submission has {set(submission_dict.keys())}, answer has {set(answer_dict.keys())}")
            return 0.0

        # Compare values
        all_match = True
        for key in answer_dict:
            submission_value = submission_dict[key]
            answer_value = answer_dict[key]

            # Try to compare as numbers
            try:
                sub_float = float(submission_value)
                ans_float = float(answer_value)
                if abs(sub_float - ans_float) >= 0.01:
                    print(f"Value mismatch for {key}: submission={sub_float}, answer={ans_float}")
                    all_match = False
                    break
            except ValueError:
                # Compare as strings (case-insensitive)
                if submission_value.lower() != answer_value.lower():
                    print(f"Value mismatch for {key}: submission='{submission_value}', answer='{answer_value}'")
                    all_match = False
                    break

        return 1.0 if all_match else 0.0

    except Exception as e:
        print(f"Error in grading: {e}")
        return 0.0



import multiprocessing
from functools import partial
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from pyquaternion import Quaternion
from tqdm import tqdm

from mlebench.grade_helpers import InvalidSubmissionError

from .mAP_evaluation import get_average_precisions, get_class_names


def parse_row(sample_token: str, prediction_string: str) -> List[Dict[str, Any]]:
    """
    Parse a row comprising of a sample_token and a prediction_string

    `prediction_string` has the form:
        `{confidence} center_x center_y center_z width length height yaw class_name`
        where `confidence` is optional

        Example row:
        ```
        97ce3ab08ccbc0baae0267cbf8d4da947e1f11ae1dbcb80c3f4408784cd9170c,1.0 2742.152625996093 673.1631800662494 -18.6561112411676 1.834 4.609 1.648 2.619835541569646 car 0.5 2728.9634555684484 657.8296521874645 -18.54676216218047 1.799 4.348 1.728 -0.5425527100619654 bus
        ```

    Returns list of dicts with keys "sample_token", "translation", "size", "rotation", "name", "score", e.g.
    ```
    [{
        'sample_token': '0f0e3ce89d2324d8b45aa55a7b4f8207fbb039a550991a5149214f98cec136ac',
        'translation': [971.8343488872263, 1713.6816097857359, -25.82534357061308],
        'size': [2.519726579986132, 7.810161372666739, 3.483438286096803],
        'rotation': [0.10913582721095375, 0.04099572636992043, 0.01927712319721745, 1.029328402625659],
        'name': 'car',
        'score': 0.3077029437237213
    }]
    ```
    """
    if prediction_string.strip() == "":
        return []
    prediction_string = prediction_string.split(" ")

    single_prediction_length = 9
    if len(prediction_string) % single_prediction_length != 0:
        raise InvalidSubmissionError(
            f"Prediction string must have {single_prediction_length} tokens per object, but got {len(prediction_string)} tokens"
        )

    n_predictions = len(prediction_string) / single_prediction_length
    preds = []
    for i in range(int(n_predictions)):
        prediction_segment = prediction_string[
            i * single_prediction_length : (i + 1) * single_prediction_length
        ]
        confidence = float(prediction_segment[0])
        prediction_segment = prediction_segment[1:]

        # Parse the prediction string
        translation = [float(x) for x in prediction_segment[0:3]]
        size = [float(x) for x in prediction_segment[3:6]]
        yaw = float(prediction_segment[6])
        name = prediction_segment[7]

        quat = Quaternion(axis=[0, 0, 1], radians=yaw)
        # (from https://www.kaggle.com/competitions/3d-object-detection-for-autonomous-vehicles/discussion/110884#638390)

        preds.append(
            {
                "sample_token": sample_token,
                "translation": translation,
                "size": size,
                "rotation": list(quat),
                "name": name,
                "score": confidence,
            }
        )
    return preds


def process_sample(gt_pred_pair, iou_thresholds):
    gt, predictions = gt_pred_pair
    class_names = get_class_names(gt)
    sample_mAPs = []
    for iou_threshold in iou_thresholds:
        average_precisions = get_average_precisions(gt, predictions, class_names, iou_threshold)
        mAP = np.mean(average_precisions)
        sample_mAPs.append(mAP)
    return sample_mAPs


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    """
    submission and answers are dataframes with columns `Id` and `PredictionString`

    We grade the submission using the 3D mAP metric described in the competition overview:
    https://www.kaggle.com/competitions/3d-object-detection-for-autonomous-vehicles/overview/evaluation

    The competition organizers provide a Python script (mAP_evaluation.py) to perform this metric calculation
    (search text for "3d version of the COCO mAP") for a (gt, predictions) pair from a single sample.
    https://www.kaggle.com/competitions/3d-object-detection-for-autonomous-vehicles/discussion/133895

    For each sample, we first parse the PredictionString in the format expected by mAP_evaluation.py, then calculate
    the mAP over the requested range of IoU thresholds, and get a mean mAP over all thresholds for that sample.

    The final score is the mean over all samples (i.e. the mean of the mean mAPs).
    """
    if len(submission) != len(answers):
        raise InvalidSubmissionError("Submission and answers must have the same length")

    if "Id" not in submission.columns:
        raise InvalidSubmissionError("Submission must have an 'Id' column")

    if "PredictionString" not in submission.columns:
        raise InvalidSubmissionError("Submission must have a 'PredictionString' column")

    assert "Id" in answers.columns, "Answers must have an 'Id' column"
    assert "PredictionString" in answers.columns, "Answers must have a 'PredictionString' column"

    submission = submission.sort_values("Id")
    answers = answers.sort_values("Id")

    if (submission["Id"].values != answers["Id"].values).any():
        raise InvalidSubmissionError("Submission and answers must have the same ids")

    # Empty values in the PredictionString column are allowed, but must be filled with an empty string
    # (pandas converts them to NaN, so we convert them back)
    submission["PredictionString"] = submission["PredictionString"].fillna("")

    # Parse each row into the format expected by mAP_evaluation.py
    submission_samples = [
        parse_row(row["Id"], row["PredictionString"]) for _, row in submission.iterrows()
    ]
    answer_samples = [
        parse_row(row["Id"], row["PredictionString"]) for _, row in answers.iterrows()
    ]

    iou_thresholds = [0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]
    # (from https://www.kaggle.com/competitions/3d-object-detection-for-autonomous-vehicles/overview/evaluation)

    # Prepare the data for parallel processing
    sample_pairs = list(zip(answer_samples, submission_samples))

    # Use multiprocessing to parallelize the computation
    num_cpus = multiprocessing.cpu_count()
    with multiprocessing.Pool(processes=num_cpus) as pool:
        results = list(
            tqdm(
                pool.imap(partial(process_sample, iou_thresholds=iou_thresholds), sample_pairs),
                total=len(sample_pairs),
                desc="Processing samples",
            )
        )

    # Flatten the results
    mAPs = [mAP for sample_mAPs in results for mAP in sample_mAPs]

    # Average over all samples and IoU thresholds
    final_mAP = np.mean(mAPs)

    return final_mAP



from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def _split_and_save(
    df: pd.DataFrame,
    test_size: float,
    public_path: Path,
    private_path: Path,
    random_state: int,
):
    """
    Helper function to perform a data split, save files, and run assertions.

    Args:
        df (pd.DataFrame): The dataframe to split.
        test_size (float): The proportion of the dataset to allocate to the test split.
        public_path (Path): The directory for public-facing files (train set, unlabeled test set).
        private_path (Path): The directory for private-facing files (labeled test set).
        random_state (int): The seed for the random number generator.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: The resulting train and test dataframes.
    """
    # Ensure output directories exist
    public_path.mkdir(parents=True, exist_ok=True)
    private_path.mkdir(parents=True, exist_ok=True)

    # Perform the split
    train_df, test_df = train_test_split(
        df, test_size=test_size, random_state=random_state
    )
    test_df_without_labels = test_df.drop(columns=["Cover_Type"])

    # Save the split data using standard filenames
    train_df.to_csv(public_path / "train.csv", index=False)
    test_df.to_csv(private_path / "test.csv", index=False)
    test_df_without_labels.to_csv(public_path / "test.csv", index=False)

    # Create a sample submission file
    submission_df = pd.DataFrame(
        {"Id": test_df["Id"].values, "Cover_Type": 2}
    )
    submission_df.to_csv(public_path / "sample_submission.csv", index=False)

    # Assertions
    assert len(train_df) + len(test_df) == len(
        df
    ), "train and test splits lengths do not add up to original data."
    assert set(train_df["Id"]).isdisjoint(
        set(test_df["Id"])
    ), f"there are Ids in both train and test sets: {set(train_df['Id']).intersection(set(test_df['Id']))}"
    assert len(test_df_without_labels) == len(
        test_df
    ), "public and Private tests should have equal length"

    return train_df, test_df


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split in public_val/private_val directories.
    """
    # Create train, test from train split
    old_train = read_csv(raw / "train.csv")

    # --- First Split: Create the original train and test sets ---
    # This call produces the original outputs, which must remain unchanged.
    new_train, new_test = _split_and_save(
        df=old_train,
        test_size=0.1,
        public_path=public,
        private_path=private,
        random_state=0,
    )

    # --- Second Split: Create a new train and validation set ---
    # Define paths for the new validation set outputs
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Calculate the test size for the second split to make the new validation
    # set (`test_val`) have the same number of samples as the original test set.
    test_val_size = len(new_test) / len(new_train)

    # This call takes the `new_train` data and splits it again, saving the
    # results to the new `_val` directories with identical structure and filenames.
    _split_and_save(
        df=new_train,
        test_size=test_val_size,
        public_path=public_val,
        private_path=private_val,
        random_state=0,  # Use the same random state for consistency
    )


import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):
    # Create train, test from train split
    old_train = read_csv(raw / "train.csv")

    # Train is c. 55M rows, original test is 9914 rows
    new_train, new_test = train_test_split(old_train, test_size=9914, random_state=0)
    new_test_without_labels = new_test.drop(columns=["fare_amount"])

    # Create a sample submission file
    submission_df = new_test.copy()[["key"]]
    submission_df["fare_amount"] = 11.35

    # Write CSVs
    new_train.to_csv(public / "labels.csv", index=False)
    new_test_without_labels.to_csv(public / "test.csv", index=False)
    submission_df.to_csv(public / "sample_submission.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)

    # Copy over other files
    shutil.copy(raw / "GCP-Coupons-Instructions.rtf", public / "GCP-Coupons-Instructions.rtf")

    # Checks
    assert set(new_train["key"]).isdisjoint(
        set(new_test["key"])
    ), "Train and test sets share samples!"
    assert new_test.shape[1] == 8, f"Test set should have 8 columns, but has {new_test.shape[1]}"
    assert (
        new_test_without_labels.shape[1] == 7
    ), f"Test set without labels should have 7 columns, but has {new_test_without_labels.shape[1]}"
    assert new_train.shape[1] == 8, f"Train set should have 8 columns, but has {new_train.shape[1]}"
    assert (
        submission_df.shape[1] == 2
    ), f"Sample submission should have 2 columns, but has {submission_df.shape[1]}"
    assert (
        submission_df.shape[0] == new_test.shape[0]
    ), f"Sample submission should have {new_test.shape[0]} rows, but has {submission_df.shape[0]}"
